#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EMBL-EBI
#
#  File author(s):
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the PSICQUIC web service

.. topic:: What is PSICQUIC ?

    :URL: http://code.google.com/p/psicquic/
    :REST: http://code.google.com/p/psicquic/wiki/PsicquicSpec_1_3_Rest

    .. highlights::

        "PSICQUIC is an effort from the HUPO Proteomics Standard Initiative
        (HUPO-PSI) to standardise the access to molecular interaction databases
        programmatically. The PSICQUIC View web interface shows that PSICQUIC
        provides access to 25 active service "

        -- Dec 2012


.. _about_queries:

About queries
================

.. rubric:: source: PSICQUIC View web page

The idea behind PSICQUIC is to retrieve information related to protein
interactions from various databases. Note that protein interactions does not
necesseraly mean protein-protein interactions. In order to be effective, the
query format has been standarised.

To do a search you can use the Molecular Interaction Query Language which is
based on Lucene's syntax. Here are some rules

* Use OR or space ' ' to search for ANY of the terms in a field
* Use AND if you want to search for those interactions where ALL of your terms are found
* Use quotes (") if you look for a specific phrase (group of terms that must
  be searched together) or terms containing special characters that may otherwise
  be interpreted by our query engine (eg. ':' in a GO term)
* Use parenthesis for complex queries (e.g. '(XXX OR YYY) AND ZZZ')
* Wildcards (`*`,?) can be used between letters in a term or at the end of terms to do fuzzy queries,
   but never at the beginning of a term.
* Optionally, you can prepend a symbol in front of your term.
    *  + (plus): include this term. Equivalent to AND. e.g. +P12345
    *  - (minus): do not include this term. Equivalent to NOT. e.g. -P12345
    *    Nothing in front of the term. Equivalent to OR. e.g. P12345
* Implicit fields are used when no field is specified (simple search). For
  instance, if you put 'P12345' in the simple query box, this will mean the same
  as identifier:P12345 OR pubid:P12345 OR pubauth:P12345 OR species:P12345 OR
  type:P12345 OR detmethod:P12345 OR interaction_id:P12345

About the MITAB output
=========================

The output returned by a query contains a list of entries. Each entry is
formatted following the MITAB output.

Here below are listed the name of the field returned ordered as they would
appear in one entry. The first item is always idA whatever version of MITAB is
used. The version 25 of MITAB contains the first 15 fields in the table below.
Newer version may incude more fields but always include the 15 from MITAB 25 in
the same order.  See the link from **irefindex**
`about mitab <http://irefindex.uio.no/wiki/README_MITAB2.6_for_iRefIndex_8.0#What_each_line_represents>`_
for more information.

=============== =========================================== =============== ======================
Field Name      Searches on                                 Implicit*       Example
=============== =========================================== =============== ======================
idA             Identifier A                                No              idA:P74565
idB             Identifier B                                No              idB:P74565
id              Identifiers (A or B)                        No              id:P74565
alias           Aliases (A or B)                            No              alias:(KHDRBS1 HCK)
identifiers     Identifiers and Aliases undistinctively     Yes             identifier:P74565
pubauth         Publication 1st author(s)                   Yes             pubauth:scott
pubid           Publication Identifier(s) OR                Yes             pubid:(10837477 12029088)
taxidA          Tax ID interactor A: the tax ID or
                the species name                            No              taxidA:mouse
taxidB          Tax ID interactor B: the tax ID or
                species name                                No              taxidB:9606
species         Species. Tax ID A or Tax ID B               Yes             species:human
type            Interaction type(s)                         Yes             type:"physical interaction"
detmethod       Interaction Detection method(s)             Yes             detmethod:"two hybrid*"
interaction_id  Interaction identifier(s)                   Yes             interaction_id:EBI-761050
pbioroleA       Biological role A                           Yes             pbioroleA:ancillary
pbioroleB       Biological role B                           Yes             pbioroleB:"MI:0684"
pbiorole        Biological roles (A or B)                   Yes             pbiorole:enzyme
ptypeA          Interactor type A                           Yes             ptypeA:protein
ptypeB          Interactor type B                           Yes             ptypeB:"gene"
ptype           Interactor types (A or B)                   Yes             pbiorole:"small molecule"
pxrefA          Interactor xref A (or Identifier A)         Yes             pxrefA:"GO:0003824"
pxrefB          Interactor xref B (or Identifier B)                         Yes pxrefB:"GO:0003824"
pxref           Interactor xrefs (A or B or Identifier
                A or Identifier B)                          Yes             pxref:"catalytic activity"
xref            Interaction xrefs (or Interaction
                identifiers)                                Yes             xref:"nuclear pore"
annot           Interaction annotations and tags            Yes             annot:"internally curated"
udate           Update date                                 Yes             udate:[20100101 TO 20120101]
negative        Negative interaction boolean                Yes             negative:true
complex         Complex expansion                           Yes             complex:"spoke expanded"
ftypeA          Feature type of participant A               Yes             ftypeA:"sufficient to bind"
ftypeB          Feature type of participant B               Yes             ftypeB:mutation
ftype           Feature type of participant A or B          Yes             ftype:"binding site"
pmethodA        Participant identification method A         Yes             pmethodA:"western blot"
pmethodB        Participant identification method B         Yes             pmethodB:"sequence tag identification"
pmethod         Participant identification methods
                 (A or B)                                   Yes             pmethod:immunostaining
stc             Stoichiometry (A or B). Only true or
                false, just to be able to filter
                interaction having stoichiometry available  Yes             stc:true
param           Interaction parameters. Only true or
                false, just to be able to filter
                interaction having parameters available     Yes             param:true
=============== =========================================== =============== ======================



"""

from bioservices import REST, UniProt


# http://code.google.com/p/psicquic/wiki/PsicquicSpec_1_3_Rest

# http://www.biocatalogue.org/services/2078#operations

__all__ = ["PSICQUIC"]


class PSICQUIC:
    """Interface to the `PSICQUIC <http://code.google.com/p/psicquic/>`_ service

    There are 2 interfaces to the PSICQUIC service (REST and WSDL) but we used
    the REST only.


    This service provides a common interface to more than 25 other services
    related to protein. So, we won't detail all the possiblity of this service.
    Here is an example that consists of looking for interactors of the
    protein ZAP70 within the IntAct database::

        >>> from bioservices import *
        >>> s = PSICQUIC()
        >>> res = s.query("intact", "zap70")
        >>> len(res) # there are 11 interactions found
        11
        >>> for x in res[1]:
        ...     print(x)
        uniprotkb:O95169
        uniprotkb:P43403
        intact:EBI-716238
        intact:EBI-1211276
        psi-mi:ndub8_human(display_long)|uniprotkb:NADH-ubiquinone oxidoreductase ASHI
        .
        .

    Here we have a list of entries. There are 15 of them (depending on
    the *output* parameter). The meaning of the entries is described on PSICQUIC
    website: https://code.google.com/p/psicquic/wiki/MITAB25Format . In short:


    #. Unique identifier for interactor A
    #. Unique identifier for interactor B.
    #. Alternative identifier for interactor A, for example the official gene
    #. Alternative identifier for interactor B.
    #. Aliases for A, separated by "|
    #. Aliases for B.
    #. Interaction detection methods, taken from the corresponding PSI-MI
    #. First author surname(s) of the publication(s)
    #. Identifier of the publication
    #. NCBI Taxonomy identifier for interactor A.
    #. NCBI Taxonomy identifier for interactor B.
    #. Interaction types,
    #. Source databases and identifiers,
    #. Interaction identifier(s) i
    #. Confidence score. Denoted as scoreType:value.

    Another example with reactome database::

        res = s.query("reactome", "Q9Y266")

    .. warning:: PSICQUIC gives access to 25 other services. We cannot create
        a dedicated parsing for all of them. So, the ::`query` method returns
        the raw data. Addition class may provide dedicated parsing in the
        future.

    .. seealso:: :class:`bioservices.biogrid.BioGRID`
    """

    _formats = [
        "tab25",
        "tab26",
        "tab27",
        "xml25",
        "count",
        "biopax",
        "xgmml",
        "rdf-xml",
        "rdf-xml-abbrev",
        "rdf-n3",
        "rdf-turtle",
    ]

    # note the typo in "genbank indentifier from bind DB
    _mapping_uniprot = {
        "genbank indentifier": "P_GI",
        "entrezgene/locuslink": "P_ENTREZGENEID",
        "uniprotkb": "ACC+ID",
        "rcsb pdb": "PDB_ID",
        "ensembl": "ENSEMBL_ID",
        "refseq": "P_REFSEQ_AC",
        "hgnc": "HGNC_ID",
        "kegg": "KEGG_ID",
        "entrez gene/locuslink": "P_ENTREZGENEID",
        "chembl": "CHEMBL_ID",
        "ddbj/embl/genbank": "EMBL_ID",
        "dip": "DIP_ID",
        "ensemblgenomes": "ENSEMBLGENOME_ID",
        "omim": "MIM_ID",
        "chebi": None,
        "chembl": None,
        #        "intact": None
    }

    # unknown: hprd, omim, bind, bind complexid, mdl,

    def __init__(self, verbose=True):
        """.. rubric:: Constructor

        :param bool verbose: print informative messages

        .. doctest::

            >>> from bioservices import PSICQUIC
            >>> s = PSICQUIC()

        """
        self.services = REST(
            "PSICQUIC",
            verbose=verbose,
            url="https://www.ebi.ac.uk/Tools/webservices/psicquic",
            url_defined_later=True,
        )  # this prevent annoying warning

        self._registry = None

        try:
            self.uniprot = UniProt(verbose=False)
        except:
            self.services.logging.warning("UniProt service could be be initialised")
        self.buffer = {}

    def _get_formats(self):
        return PSICQUIC._formats

    formats = property(_get_formats, doc="Returns the possible output formats")

    def _get_active_db(self):
        names = self.registry_names[:]
        actives = self.registry_actives[:]
        names = [x.lower() for x, y in zip(names, actives) if y == "true"]
        return names

    activeDBs = property(_get_active_db, doc="returns the active DBs only")

    def read_registry(self):
        """Reads and returns the active registry"""
        url = "registry/registry?action=ACTIVE&format=txt"
        res = self.services.http_get(url, frmt="txt")
        return res.split()

    def print_status(self):
        """Prints the services that are available

        :return: Nothing

        The output is tabulated. The columns are:

        * names
        * active
        * count
        * version
        * rest URL
        * soap URL
        * rest example
        * restricted

        .. seealso:: If you want the data into lists, see all attributes
            starting with registry such as :meth:`registry_names`
        """
        url = "registry/registry?action=STATUS&format=xml"
        res = self.services.http_get(url, frmt="txt")

        names = self.registry_names
        counts = self.registry_counts
        versions = self.registry_versions
        actives = self.registry_actives
        resturls = self.registry_resturls
        soapurls = self.registry_soapurls
        restexs = self.registry_restexamples
        restricted = self.registry_restricted
        N = len(names)

        indices = sorted(range(0, N), key=lambda k: names[k])

        for i in range(0, N):
            print(
                "%s\t %s\t %s\t %s\t %s %s %s %s\n"
                % (
                    names[i],
                    actives[i],
                    counts[i],
                    versions[i],
                    resturls[i],
                    soapurls[i],
                    restexs[i],
                    restricted[i],
                )
            )

    # todo a property for the version of PISCQUIC

    def _get_registry(self):
        if self._registry is None:
            url = "registry/registry?action=STATUS&format=xml"
            res = self.services.http_get(url, frmt="xml")
            res = self.services.easyXML(res)
            self._registry = res
        return self._registry

    registry = property(_get_registry, doc="returns the registry of psicquic")

    def _get_registry_names(self):
        res = self.registry
        return [x.findAll("name")[0].text for x in res.findAll("service")]

    registry_names = property(_get_registry_names, doc="returns all services available (names)")

    def _get_registry_restricted(self):
        res = self.registry
        return [x.findAll("restricted")[0].text for x in res.findAll("service")]

    registry_restricted = property(_get_registry_restricted, doc="returns restricted status of services")

    def _get_registry_resturl(self):
        res = self.registry
        data = [x.findAll("resturl")[0].text for x in res.findAll("service")]
        return data

    registry_resturls = property(_get_registry_resturl, doc="returns URL of REST services")

    def _get_registry_restex(self):
        res = self.registry
        data = [x.findAll("restexample")[0].text for x in res.findAll("service")]
        return data

    registry_restexamples = property(_get_registry_restex, doc="retuns REST example for each service")

    def _get_registry_soapurl(self):
        res = self.registry
        return [x.findAll("soapurl")[0].text for x in res.findAll("service")]

    registry_soapurls = property(_get_registry_soapurl, doc="returns URL of WSDL service")

    def _get_registry_active(self):
        res = self.registry
        return [x.findAll("active")[0].text for x in res.findAll("service")]

    registry_actives = property(_get_registry_active, doc="returns active state of each service")

    def _get_registry_count(self):
        res = self.registry
        return [x.findAll("count")[0].text for x in res.findAll("service")]

    registry_counts = property(_get_registry_count, doc="returns number of entries in each service")

    def _get_registry_version(self):
        res = self.registry
        names = [x.findAll("name")[0].text for x in res.findAll("service")]
        N = len(names)
        version = [0] * N
        for i in range(0, N):
            x = res.findAll("service")[i]
            if x.findAll("version"):
                version[i] = x.findAll("version")[0].text
            else:
                version[i] = None
        return version

    registry_versions = property(_get_registry_version, doc="returns version of each service")

    def query(
        self,
        service,
        query,
        output="tab25",
        version="current",
        firstResult=None,
        maxResults=None,
    ):
        """Send a query to a specific database

        :param str service: a registered service. See :attr:`registry_names`.
        :param str query: a valid query. Can be `*` or a protein name.
        :param str output: a valid format. See s._formats

        ::

            s.query("intact", "brca2", "tab27")
            s.query("intact", "zap70", "xml25")
            s.query("matrixdb", "*", "xml25")

        This is the programmatic approach to this website:

        http://www.ebi.ac.uk/Tools/webservices/psicquic/view/main.xhtml


        Another example consist in accessing the *string* database for fetching
        protein-protein interaction data of a particular model organism. Here we
        restrict the query to 100 results::

            s.query("string", "species:10090", firstResult=0, maxResults=100, output="tab25")

        # spaces are automatically converted

            s.query("biogrid", "ZAP70 AND species:9606")

        .. warning:: AND must be in big caps. Some database are ore permissive
            than other (e.g., intact accepts "and"). species must be a valid ID number. Again, some DB are more
            permissive and may accept the name (e.g., human)

        To obtain the number of interactions in intact for the human specy::

            >>> len(p.query("intact", "species:9606"))


        """
        if service not in self.activeDBs:
            raise ValueError("database %s not in active databases" % service)

        params = {}
        if output is not None:
            self.services.devtools.check_param_in_list(output, self.formats)
            params["format"] = output
        else:
            output = "none"

        names = [x.lower() for x in self.registry_names]
        try:
            index = names.index(service)
        except ValueError:
            self.logging.error("The service you gave (%s) is not registered. See self.registery_names" % service)
            raise ValueError

        # get the base url according to the service requested
        resturl = self.registry_resturls[index]

        if firstResult is not None:
            params["firstResult"] = firstResult
        if maxResults is not None:
            params["maxResults"] = maxResults

        url = resturl + "query/" + query

        if "xml" in output:
            res = self.services.http_get(url, frmt="xml", params=params)
        else:
            res = self.services.http_get(url, frmt="txt", params=params)
            res = res.strip().split("\n")

        if output.startswith("tab"):
            res = self._convert_tab2dict(res)

        return res

    def _convert_tab2dict(self, data):
        """

        https://code.google.com/p/psicquic/wiki/MITAB26Format
        """
        results = []
        for line in data:
            results.append(line.split("\t"))

        return results

    def queryAll(
        self,
        query,
        databases=None,
        output="tab25",
        version="current",
        firstResult=None,
        maxResults=None,
    ):
        """Same as query but runs on all active database

        :param list databases: database to query. Queries all active DB if not provided
        :return: dictionary where keys correspond to databases and values to the output of the query.

        ::

            res = s.queryAll("ZAP70 AND species:9606")
        """

        results = {}
        if databases is None:
            databases = [x.lower() for x in self.activeDBs]

        for x in databases:
            if x not in self.activeDBs:
                raise ValueError("database %s not in active databases" % x)

        for name in databases:
            self.logging.warning("Querying %s" % name),
            res = self.query(
                name,
                query,
                output=output,
                version=version,
                firstResult=firstResult,
                maxResults=maxResults,
            )
            if output.startswith("tab25"):
                results[name] = [x for x in res if x != [""]]
            else:
                import copy

                results[name] = copy.copy(res)
        for name in databases:
            self.logging.info("Found %s in %s" % (len(results[name]), name))
        return results

    def getInteractionCounter(self, query):
        """Returns a dictionary with database as key and results as values

        :param str query: a valid query
        :return: a dictionary which key as database and value as number of entries

        Consider only the active database.

        """
        # get the active names only
        activeDBs = self.activeDBs[:]
        res = [(str(name), int(self.query(name, query, output="count")[0])) for name in activeDBs]
        return dict(res)

    def getName(self, data):
        idsA = [x[0] for x in data]
        idsB = [x[1] for x in data]
        return idsA, idsB

    def knownName(self, data):
        """Scan all entries (MITAB) and returns simplified version


        Each item in the input list of mitab entry
        The output is made of 2 lists corresponding to
        interactor A and B found in the mitab entries.

        elements in the input list takes the following forms::

            DB1:ID1|DB2:ID2
            DB3:ID3

        The | sign separates equivalent IDs from different databases.

        We want to keep only one. The first known databae is kept. If in the list of DB:ID pairs no known
        database is found, then we keep the first one whatsover.

        known databases are those available in the uniprot mapping tools.

        chembl and chebi IDs are kept unchanged.


        """
        self.logging.info("converting data into known names")
        idsA = [x[0].replace('"', "") for x in data]
        idsB = [x[1].replace('"', "") for x in data]
        # extract the first and second ID but let us check if it is part of a
        # known uniprot mapping.Otherwise no conversion will be possible.
        # If so, we set the ID to "unknown"
        # remove the " character that can be found in a few cases (e.g,
        # chebi:"CHEBI:29036")
        # idsA = [x.replace("chebi:CHEBI:","chebi:") for x in idsA]
        # idsB = [x.replace("chebi:CHEBI:", "chebi:") for x in idsB]

        # special case:
        # in mint, there is an entry that ends with a | uniprotkb:P17844|
        idsA = [x.strip("|") for x in idsA]
        idsB = [x.strip("|") for x in idsB]

        # the first ID
        for i, entry in enumerate(idsA):
            try:
                dbs = [x.split(":")[0] for x in entry.split("|")]
                IDs = [x.split(":")[1] for x in entry.split("|")]
                valid_dbs = [(db, ID) for db, ID in zip(dbs, IDs) if db in self._mapping_uniprot.keys()]
                # search for an existing DB
                if len(valid_dbs) >= 1:
                    idsA[i] = valid_dbs[0][0] + ":" + valid_dbs[0][1]
                else:
                    self.logging.debug("none of the DB for this entry (%s) are available" % (entry))
                    idsA[i] = "?" + dbs[0] + ":" + IDs[0]
            except:
                self.logging.info("Could not extract name from %s" % entry)
                idsA[i] = "??:" + entry  # we add a : so that we are sure that a split(":") will work
        # the second ID
        for i, entry in enumerate(idsB):
            try:
                dbs = [x.split(":")[0] for x in entry.split("|")]
                IDs = [x.split(":")[1] for x in entry.split("|")]
                valid_dbs = [(db, ID) for db, ID in zip(dbs, IDs) if db in self._mapping_uniprot.keys()]
                # search for an existing DB
                if len(valid_dbs) >= 1:
                    idsB[i] = valid_dbs[0][0] + ":" + valid_dbs[0][1]
                else:
                    self.logging.debug("none of the DB (%s) for this entry are available" % (entry))
                    idsB[i] = "?" + dbs[0] + ":" + IDs[0]
            except:
                self.logging.info("Could not extract name from %s" % entry)
                idsB[i] = "??:" + entry

        countA = len([x for x in idsA if x.startswith("?")])
        countB = len([x for x in idsB if x.startswith("?")])
        if countA + countB > 0:
            self.logging.warning("%s ids out of %s were not identified" % (countA + countB, len(idsA) * 2))
            print(set([x.split(":")[0] for x in idsA if x.startswith("?")]))
            print(set([x.split(":")[0] for x in idsB if x.startswith("?")]))
        self.logging.info("knownName done")
        return idsA, idsB

    def preCleaning(self, data):
        """remove entries ehre IdA or IdB is set to "-" """
        ret = [x for x in data if x[0] != "-" and x[1] != "-"]
        return ret

    def postCleaningAll(self, data, keep_only="HUMAN", flatten=True, verbose=True):
        """

        even more cleaing by ignoring score, db and interaction
        len(set([(x[0],x[1]) for x in retnew]))
        """
        results = {}
        for k in data.keys():
            self.logging.info("Post cleaning %s" % k)
            ret = self.postCleaning(data[k], keep_only="HUMAN", verbose=verbose)
            if len(ret):
                results[k] = ret
        if flatten:
            results = [x for k in results.keys() for x in results[k]]
        return results

    def postCleaning(
        self,
        data,
        keep_only="HUMAN",
        remove_db=["chebi", "chembl"],
        keep_self_loop=False,
        verbose=True,
    ):
        """Remove entries with a None and keep only those with the keep pattern"""
        if verbose:
            print("Before removing anything: ", len(data))

        data = [x for x in data if x[0] is not None and x[1] is not None]
        if verbose:
            print("After removing the None: ", len(data))

        data = [x for x in data if x[0].startswith("!") is False and x[1].startswith("!") is False]
        if verbose:
            print("After removing the !: ", len(data))

        for db in remove_db:
            data = [x for x in data if x[0].startswith(db) is False]
            data = [x for x in data if x[1].startswith(db) is False]
            if verbose:
                print("After removing entries that match %s : " % db, len(data))

        data = [x for x in data if keep_only in x[0] and keep_only in x[1]]
        if verbose:
            print("After removing entries that don't match %s : " % keep_only, len(data))

        if keep_self_loop is False:
            data = [x for x in data if x[0] != x[1]]
            if verbose:
                print("After removing self loop : ", len(data))

        data = list(set(data))
        if verbose:
            print("After removing identical entries", len(data))

        return data

    def convertAll(self, data):
        results = {}
        for k in data.keys():
            self.logging.info("Analysing %s" % k)
            results[k] = self.convert(data[k], db=k)
        return results

    def convert(self, data, db=None):
        self.logging.debug("converting the database %s" % db)
        idsA, idsB = self.knownName(data)
        mapping = self.mappingOneDB(data)
        results = []
        for i, entry in enumerate(data):
            x = idsA[i].split(":", 1)[1]
            y = idsB[i].split(":", 1)[1]
            xp = mapping[x]
            yp = mapping[y]
            try:
                ref = entry[8]
            except:
                ref = "?"
            try:
                score = entry[14]
            except:
                score = "?"
            try:
                interaction = entry[11]
            except:
                interaction = "?"
            results.append((xp, yp, score, interaction, ref, db))
        return results

    def mappingOneDB(self, data):
        query = {}
        self.logging.debug("converting IDs with proper DB name (knownName function)")
        entriesA, entriesB = self.knownName(data)  # idsA and B contains list of a single identifier of the form db:id
        # the db is known from _mapping.uniprot otherwise it is called "unknown"

        # get unique DBs to build the query dictionary
        dbsA = [x.split(":")[0] for x in entriesA]
        dbsB = [x.split(":")[0] for x in entriesB]
        for x in set(dbsA):
            query[x] = set()

        for x in set(dbsB):
            query[x] = set()

        #
        query_copy = query.copy()
        for k in query_copy.keys():
            if k.startswith("?"):
                del query[k]

        # the data to store
        mapping = {}
        N = len(data)

        # scan all entries
        counter = 0
        for entryA, entryB in zip(entriesA, entriesB):
            counter += 1
            dbA, idA = entryA.split(":")
            try:
                dbB, idB = entryB.split(":")
            except:
                print(entryB)
            if idA not in mapping.keys():
                if dbA.startswith("?"):
                    mapping[idA] = entryA
                else:
                    query[dbA].add(idA)
            if idB not in mapping.keys():
                if dbB.startswith("?"):
                    mapping[idB] = entryB
                else:
                    query[dbB].add(idB)

            for k in query.keys():
                if len(query[k]) > 2000 or counter == N:
                    this_query = list(query[k])
                    DBname = self._mapping_uniprot[k]

                    if DBname is not None:
                        self.logging.warning("Request sent to uniprot for %s database (%s/%s)" % (DBname, counter, N))
                        res = self.uniprot.mapping(fr=DBname, to="ID", query=" ".join(this_query))
                        for x in this_query:
                            if x not in res:  # was not found
                                mapping[x] = "!" + k + ":" + x
                            else:
                                # we should be here since the queries are populated
                                # if not already in the mapping dictionary
                                if x not in res.keys():
                                    raise ValueError(x)
                                if len(res[x]) == 1:
                                    mapping[x] = res[x][0]
                                else:
                                    self.logging.warning("psicquic mapping found more than 1 id. keep first one")
                                    mapping[x] = res[x][0]
                    else:
                        for x in this_query:
                            mapping[x] = k + ":" + x
                    query[k] = set()

        for k in query.keys():
            assert len(query[k]) == 0
        return mapping


class AppsPPI(object):
    """This is an application based on PPI that search for relevant interactions

    Interctions between proteins may have a score provided by each database.
    However, scores are sometimes ommited. Besides, they may have different
    meaning for different databases. Another way to score an interaction is to
    count in how many database it is found.

    This class works as follows. First, you query a protein:


        p = AppsPPI()
        p.query("ZAP70 AND species:9606")

    This, is going to call the PSICQUIC queryAll method to send this query to
    all active databases. Then, it calls the convertAll functions to convert all
    interactors names into uniprot name if possible. If not, interactions are
    not taken into account. Finally, it removes duplicated and performs some
    cleaning inside the postCleaningall method.

    Then, you can call the summary method that counts the interactions. The
    count is stored in the attrbiute relevant_interactions.

        p.summary()

    Let us see how many intercations where found with. THe number of databases
    that contains at least one interactions is

        >>> p.N
        >>> p.relevant_interactions[N]
        [['ZAP70_HUMAN', 'DBNL_HUMAN']]

    So, there was 1 interaction found in all databases.

    """

    def __init__(self, verbose=False):
        """.. rubric:: constructor"""
        self.psicquic = PSICQUIC(verbose=False)
        self.verbose = verbose

    def queryAll(self, query, databases=None):
        """

        :param str query: a valid query. See :class:`~bioservices.psicquic.PSICQUIC`
            and :mod:`bioservices.psicquic` documentation.
        :param str databases: by default, queries are sent to each active database.
            you can overwrite this behavious by providing your own list of
            databses
        :return: nothing but the interactions attributes is populated with a
            dictionary where keys correspond to each database that returned a non empty list
            of interactions. The item for each key is a list of interactions containing the
            interactors A and B, the score, the type of intercations and the score.


        """
        # self.results_query = self.psicquic.queryAll("ZAP70 AND species:9606")
        print("Requests sent to psicquic. Can take a while, please be patient...")
        self.results_query = self.psicquic.queryAll(query, databases)
        self.interactions = self.psicquic.convertAll(self.results_query)
        self.interactions = self.psicquic.postCleaningAll(self.interactions, flatten=False, verbose=self.verbose)
        self.N = len(self.interactions.keys())
        self.counter = {}
        self.relevant_interactions = {}

    def summary(self):
        """Build some summary related to the found interactions from queryAll

        :return: nothing but the relevant_interactions and counter attribute

            p = AppsPPI()
            p.queryAll("ZAP70 AND species:9606")
            p.summary()

        """
        for k, v in self.interactions.items():
            print("Found %s interactions within %s database" % (len(v), k))

        counter = {}
        for k in self.interactions.keys():
            # scan each dabase
            for v in self.interactions[k]:
                interaction = v[0] + "++" + v[1]
                db = v[5]
                if interaction in counter.keys():
                    counter[interaction].append(db)
                else:
                    counter[interaction] = [db]
        for k in counter.keys():
            counter[k] = list(set(counter[k]))

        N = len(self.interactions.keys())

        print("-------------")
        summ = {}
        for i in range(1, N + 1):
            res = [(x.split("++"), counter[x]) for x in counter.keys() if len(counter[x]) == i]
            print("Found %s interactions in %s common databases" % (len(res), i))
            res = [x.split("++") for x in counter.keys() if len(counter[x]) == i]
            if len(res):
                summ[i] = [x for x in res]
            else:
                summ[i] = []

        self.counter = counter.copy()
        self.relevant_interactions = summ.copy()

    def get_reference(self, idA, idB):
        key = idA + "++" + idB
        uniq = len(self.counter[key])
        ret = [x for k in self.interactions.keys() for x in self.interactions[k] if x[0] == idA and x[1] == idB]
        N = len(ret)
        print("Interactions %s -- %s has %s entries in %s databases (%s):" % (idA, idB, N, uniq, self.counter[key]))
        for r in ret:
            print(r[5], " reference", r[4])

    def show_pie(self):
        """a simple example to demonstrate how to visualise number of
        interactions found in various databases

        """
        try:
            from pylab import pie, clf, title, show, legend
        except ImportError:
            from bioservices import BioServicesError

            raise BioServicesError("You must install pylab/matplotlib to use this functionality")
        labels = range(1, self.N + 1)
        print(labels)
        counting = [len(self.relevant_interactions[i]) for i in labels]

        clf()
        # pie(counting, labels=[str(int(x)) for x in labels], shadow=True)
        pie(counting, labels=[str(x) for x in counting], shadow=True)
        title("Number of interactions found in N databases")
        legend([str(x) + " database(s)" for x in labels])
        show()



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the UniProt web service

.. topic:: What is UniProt ?

    :URL: http://www.uniprot.org
    :Citation:

    .. highlights::

        "The Universal Protein Resource (UniProt) is a comprehensive resource for protein
        sequence and annotation data. The UniProt databases are the UniProt
        Knowledgebase (UniProtKB), the UniProt Reference Clusters (UniRef), and the
        UniProt Archive (UniParc). The UniProt Metagenomic and Environmental Sequences
        (UniMES) database is a repository specifically developed for metagenomic and
        environmental data."

        -- From Uniprot web site (help/about) , Dec 2012


.. mapping between uniprot and bunch of other DBs.
.. ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/
.. http://www.uniprot.org/docs/speclist
.. http://www.uniprot.org/docs/pkinfam

"""
import io
import time
import urllib
import json

import pandas as pd

from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["UniProt"]


class UniProt:
    """Interface to the `UniProt <http://www.uniprot.org>`_ service

    ::

        >>> from bioservices import UniProt
        >>> u = UniProt(verbose=False)
        >>> u.mapping("UniProtKB_AC-ID", "KEGG", query='P43403')
        defaultdict(<type 'list'>, {'P43403': ['hsa:7535']})
        >>> res = u.search("P43403")

        # Returns sequence on the ZAP70_HUMAN accession Id
        >>> sequence = u.search("ZAP70_HUMAN", columns="sequence")


    .. versionchanged:: 1.10 
    
        Uniprot update its service in June 2022. Changes were made in the bioservices
        API with small changes. User API is more or less the same. Main issues that may 
        be faced are related to change of output column names. Please see the 
        :attr:`_legacy_names` for corresponding changes.

        Some notes about searches. The *and* and *or* are now upper cases. 
        The *organism* and *taxonomy* fields are now *organism_id* and *taxonomy_id*


    """

    # June 2022, API changes and these labels changed:
    _legacy_names = {
        'id': 'accession',
        'entry name': 'id',
        'genes': 'gene_names',
        'genes(PREFERRED)':	'gene_primary',
        'genes(ALTERNATIVE)': 'gene_synonym',
        'genes(OLN)': 'gene_oln',
        'genes(ORF)':	'gene_orf',
        'organism':	'organism_name',
        'organism-id':	'organism_id',
        'protein names':	'protein_name',
        'proteome':	'xref_proteomes',
        'lineage(ALL)':	'lineage',
        'virus hosts':	'virus_hosts',

        'comment(ALTERNATIVE PRODUCTS)': 'cc_alternative_products',
        'feature(ALTERNATIVE SEQUENCE)': 'ft_var_seq',
        'comment(ERRONEOUS GENE MODEL PREDICTION)': 'error_gmodel_pred',
        'fragment': 'fragment',
        'encodedon': 'organelle',
        'length': 'length',
        'mass': 'mass',
        'comment(MASS SPECTROMETRY)': 'cc_mass_spectrometry',
        'feature(NATURAL VARIANT)': 'ft_variant',
        'feature(NON ADJACENT RESIDUES)': 'ft_non_cons',
        'feature(NON STANDARD RESIDUE)': 'ft_non_std',
        'feature(NON TERMINAL RESIDUE)': 'ft_non_ter',
        'comment(POLYMORPHISM)': 'cc_polymorphism',
        'comment(RNA EDITING)': 'cc_rna_editing',
        'sequence': 'sequence',
        'comment(SEQUENCE CAUTION)': 'cc_sequence_caution',
        'feature(SEQUENCE CONFLICT)': 'ft_conflict',
        'feature(SEQUENCE UNCERTAINTY)': 'ft_unsure',
        'version(sequence)': 'sequence_version',

        # function
        'comment(ABSORPTION)': 'absorption',
        'feature(ACTIVE SITE)': 'ft_act_site',
        'comment(ACTIVITY REGULATION)': 'cc_activity_regulation',
        'feature(BINDING SITE)': 'ft_binding',
        'chebi': 'ft_ca_bind',
        'chebi(Catalytic activity)': 'cc_catalytic_activity',
        'chebi(Cofactor)': 'cc_cofactor',
        'feature(DNA BINDING)': 'ft_dna_bind',
        'ec': 'ec',
        'comment(FUNCTION)': 'cc_function',
        'comment(KINETICS)': 'kinetics',
        'feature(METAL BINDING)': 'ft_metal',
        'feature(NP BIND)': 'ft_np_bind',
        'comment(PATHWAY)': 'cc_pathway',
        'comment(PH DEPENDENCE)': 'ph_dependence',
        'comment(REDOX POTENTIAL)': 'redox_potential',
        'rhea-id': 'rhea_id',
        'feature(SITE)': 'ft_site',
        'comment(TEMPERATURE DEPENDENCE)': 'temp_dependence',

        # misc
        'annotation score': 'annotation_score',
        'comment(CAUTION)': 'cc_caution',
        'features': 'feature',
        'keyword-id': 'keywordid',
        'keywords': 'keyword',
        'comment(MISCELLANEOUS)': 'cc_miscellaneous',
        'existence': 'protein_existence',
        'reviewed': 'reviewed',
        'tools': 'tools',
        'uniparcid': 'uniparc_id',

        # Interaction =============================
        "interactor": "cc_interaction",
        "comment(SUBUNIT)": "cc_subunit",

        # GO
        "go": "go",
        "go(biological process)": "go_p",
        "go(cellular component)": "go_c",
        "go(molecular function)": "go_f",
        "go-id": "go_id",

        # Date of
        "created": "date_created",
        "last-modified": "date_modified",
        "sequence-modified": "date_sequence_modified",
        "version(entry)": "version",
        # STRUCTURE
        "3d": "structure_3d",
        "feature(BETA STRAND)": "ft_strand",
        "feature(HELIX)": "ft_helix",
        "feature(TURN)": "ft_turn",

        # subcellular function
        "comment(SUBCELLULAR LOCATION)":"cc_subcellular_location",
        "feature(INTRAMEMBRANE)":"ft_intramem",
        "feature(TOPOLOGICAL DOMAIN)":"ft_topo_dom",
        "feature(TRANSMEMBRANE)": "ft_transmem",

        # Pathology
        'comment(ALLERGEN)': 'cc_allergen',
        'comment(BIOTECHNOLOGY)': 'cc_biotechnology',
        'comment(DISRUPTION PHENOTYPE)': 'cc_disruption_phenotype',
        'comment(DISEASE)': 'cc_disease',
        'feature(MUTAGENESIS)': 'ft_mutagen',
        'comment(PHARMACEUTICAL)': 'cc_pharmaceutical',
        'comment(TOXIC DOSE)': 'cc_toxic_dose',

        # PTM
        'feature(CHAIN)': 'ft_chain',
        'feature(CROSS LINK)': 'ft_crosslnk',
        'feature(DISULFIDE BOND)': 'ft_disulfid',
        'feature(GLYCOSYLATION)': 'ft_carbohyd',
        'feature(INITIATOR METHIONINE)': 'ft_init_met',
        'feature(LIPIDATION)': 'ft_lipid',
        'feature(MODIFIED RESIDUE)': 'ft_mod_res',
        'feature(PEPTIDE)': 'ft_peptide',
        'comment(PTM)': 'cc_ptm',
        'feature(PROPEPTIDE)': 'ft_propep',
        'feature(SIGNAL)': 'ft_signal',
        'feature(TRANSIT)': 'ft_transit',

        # Family domains
        'feature(COILED COIL)': 'ft_coiled',
        'feature(COMPOSITIONAL BIAS)': 'ft_compbias',
        'comment(DOMAIN)': 'cc_domain',
        'feature(DOMAIN EXTENT)': 'ft_domain',
        'feature(MOTIF)': 'ft_motif',
        'families': 'protein_families',
        'feature(REGION)': 'ft_region',
        'feature(REPEAT)': 'ft_repeat',
        'comment(SIMILARITY)': '<does not exist>',
        'feature(ZINC FINGER)': 'ft_zn_fing',

    }

    _valid_columns = [
        # Names & Taxonomy ================================================
        "accession",
        "id",
        "gene_names",
        "gene_primary",
        "gene_synonym",
        "gene_oln",
        "gene_orf",
        "organism_name",
        "organism_id",
        "protein_name",
        "xref_proteomes",
        "lineage",
        "virus_hosts",

        # Sequences ========================================================
        "fragment",
        "sequence",
        "length",
        "mass",
        "organelle",
        "cc_alternative_products",
        "error_gmodel_pred",
        "cc_mass_spectrometry",
        "cc_polymorphism",
        "cc_rna_editing",
        "cc_sequence_caution",
        "ft_var_seq",
        "ft_variant",
        "ft_non_cons",
        "ft_non_std",
        "ft_non_ter",
        "ft_conflict",
        "ft_unsure",
        "sequence_version",
        
        # Family and Domains ========================================
        'ft_coiled',
        'ft_compbias',
        'cc_domain',
        'ft_domain',
        'ft_motif',
        'protein_families',
        'ft_region',
        'ft_repeat',
        'ft_zn_fing',

        # Function ===================================================
        'absorption',
        'ft_act_site',
        'cc_activity_regulation',
        'ft_binding',
        #'ft_ca_bind',
        'cc_catalytic_activity',
        'cc_cofactor',
        'ft_dna_bind',
        'ec',
        'cc_function',
        'kinetics',
        #'ft_metal',
        #'ft_np_bind',
        'cc_pathway',
        'ph_dependence',
        'redox_potential',
        #'rhea_id',
        'ft_site',
        'temp_dependence',

        # Gene Ontology ==================================
        "go",
        "go_p",
        "go_f",
        "go_c",
        "go_id",

        # Interaction ======================================
        "cc_interaction",
        "cc_subunit",

        # EXPRESSION =======================================
        "cc_developmental_stage",
        "cc_induction",
        "cc_tissue_specificity",

        # Publications
        "lit_pubmed_id",
        
        # Date of
        "date_created",
        "date_modified",
        "date_sequence_modified",
        "version",

        # Structure
        "structure_3d",
        "ft_strand",
        "ft_helix",
        "ft_turn",

        # Subcellular location
        "cc_subcellular_location",
        "ft_intramem",
        "ft_topo_dom",
        "ft_transmem",

        # Miscellaneous ==========================
        "annotation_score",
        "cc_caution",
        "comment_count",
        #"feature",
        "feature_count",
        "keyword",
        "keywordid",
        "cc_miscellaneous",
        "protein_existence",
        "tools",
        "reviewed",
        "uniparc_id",
        
        # Pathology
        'cc_allergen',
        'cc_biotechnology',
        'cc_disruption_phenotype',
        'cc_disease',
        'ft_mutagen',
        'cc_pharmaceutical',
        'cc_toxic_dose',

        # PTM / Processsing
        'ft_chain',
        'ft_crosslnk',
        'ft_disulfid',
        'ft_carbohyd',
        'ft_init_met',
        'ft_lipid',
        'ft_mod_res',
        'ft_peptide',
        'cc_ptm',
        'ft_propep',
        'ft_signal',
        'ft_transit',

        # not documented
        'xref_pdb'
    ]
    _url = "https://rest.uniprot.org"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        :param cache: set to True to cache request
        """

        self.services = REST(name="UniProt", url=UniProt._url, verbose=verbose, cache=cache, url_defined_later=True)

        self.TIMEOUT = 100
        self._valid_mapping = None
        self._database = "uniprot"

    def _download_flat_files(self, output="uniprot_sprot.dat.gz"): #pragma: no cover
        """could be used to get all data in flat files (about compressed 500Mb )"""
        # deprecated in v1.10 due to API change in uniprot
        url = "ftp://ftp.ebi.ac.uk/pub/databases/uniprot/knowledgebase/uniprot_sprot.dat.gz"
        self.services.logging.info("Downloading uniprot file from the web. May take some time.:")
        urllib.request.urlretrieve(url, output)

    def _get_valid_mapping(self):
        if not self._valid_mapping:
            self._set_valid_mapping()
        return self._valid_mapping

    def _set_valid_mapping(self):
        fields = self.services.http_get("configure/idmapping/fields")
        groups = fields["groups"]
        rules = {}
        for item in fields["rules"]:
            ID = item['ruleId']
            rules[ID] = item

        # This is suppose to be a set of database name available in Uniprot 
        from_to = {}
        for item in [x for group in groups for x in group['items']]:
            # should be name, not DisplayName
            name = item['name']
            if item['from']:
                tos = rules[item['ruleId']]['tos']
                from_to[name] = tos

        self._valid_mapping = from_to
    
    valid_mapping = property(_get_valid_mapping, _set_valid_mapping)

    def mapping(self, fr="UniProtKB_AC-ID", to="KEGG", query="P13368", polling_interval_seconds=3, max_waiting_time=100):
        """This is an interface to the UniProt mapping service

        :param fr: the source database identifier. See :attr:`valid_mapping`.
        :param to: the targetted database identifier. See :attr:`valid_mapping`.
        :param query: a string containing one or more IDs separated by a space
            It can also be a list of strings.
        :param polling_interval_seconds: the number of seconds between each status check of the current job
        :param max_waiting_time: the maximum number of seconds to wait for the final answer.
        :return: a dictionary with two possible keys. The first one is 'results' 
            with the from / to answers and the second one 'failedIds' with Ids that were not found

        ::

            >>> u.mapping("UniProtKB_AC-ID", "KEGG", 'P43403')
            {'results': [{'from': 'P43403', 'to': 'hsa:7535'}]}

        The output is a dictionary. Identifiers that were not found are stored in the keys 
        'failedIds'. Succesful queries are stored in the 'results' key that is a list
        of dictionaries with two keys set to 'from' and 'to'. The 'from' key should be in your input list.
        The 'to' key is the result. Here we have the KEGG identifier recognised by its prefix 'hsa:', which is for human. 
        Sometimes the output ('to') it is more complicated. Consider the following  example::

            u.mapping("UniParc", "UniProtKB", 'UPI0000000001,UPI0000000002')

        You will see that the UniParc results is more complex than just an identifier.

        See :attr:`valid_mapping` attribut for list of valid mapping identifiers.

        Note that according to Uniprot (June 2022), there are various limits on ID Mapping Job Submission:

        ========= =====================================================================================
        Limit	  Details     
        ========= =====================================================================================
        100,000	  Total number of ids allowed in comma separated param ids in /idmapping/run api
        500,000	  Total number of "mapped to" ids allowed
        100,000	  Total number of "mapped to" ids allowed to be enriched by UniProt data
        10,000	  Total number of "mapped to" ids allowed with filtering
        ========= =====================================================================================

        .. versionchanged:: 1.1.1 to return a dictionary instaed of a list
        .. versionchanged:: 1.1.2 the values for each key is now made of a list
            instead of strings so as to store more than one values.
        .. versionchanged:: 1.2.0 input query can also be a list of strings
            instead of just a string
        .. versionchanged:: 1.3.1:: use http_post instead of http_get. This is 3 times
            faster and allows queries with more than 600 entries in one go.
        .. version 1.10.0:: new API due to  uniprot website update
        """

        if isinstance(query, (list, tuple)):
            query = ",".join(query)
        elif isinstance(query,str):
            pass

        # First, we call the real mapping request
        params = {"from": fr, "to": to, "ids": query}

        job = self.services.http_post("idmapping/run", frmt="json", data=params)
        try:
            job_id = job['jobId']
        except TypeError:
             logger.error(self.services.last_response.content.decode())
             return

        # the job id will tell us about the job status 
        results = None
        waiting_time = 0
        while not results and waiting_time < max_waiting_time:   
            logger.info("Waiting for {job_id} to complete")         
            results = self.services.http_get(f"idmapping/status/{job_id}", frmt="json")
            if results != 500 and 'results' in results:
                return results
            else: #pragma: no cover
                time.sleep(polling_interval_seconds)
                results = None
            waiting_time += polling_interval_seconds

    def retrieve(self, uniprot_id, frmt="json", database="uniprot", include=False):
        """Search for a uniprot ID in UniProtKB database

        :param str uniprot: a valid UniProtKB ID, or uniref, uniparc or taxonomy. 
        :param str frmt: expected output format amongst xml, txt, fasta, gff, rdf
        :param str database:  database name in (uniprot, uniparc, uniref, taxonomy)
        :param bool include: include data with RDF format.
        :return: if the parameter uniprot_id is string, the output will be a a list of identifiers is provided, the output is also a list
            otherwise, a string. The content of the string of items in the list
            depends on the value of **frmt**.

        ::

            >>> u = UniProt()
            >>> res = u.retrieve("P09958", frmt="txt")
            >>> fasta = u.retrieve(['P29317', 'Q5BKX8', 'Q8TCD6'], frmt='fasta')
            >>> print(fasta[0])


        .. versionchanged:: 1.10 the xml format is now returned as raw XML. It is not
            interpreted anymore. The RDF has now an additional option to include data 
            from referenced data sets directly in the returned data (set include=True parameter).
            Default output format is now set to json. 
        """
        if database == 'uniprot': 
            if frmt not in ("txt", "xml", "rdf", "gff", "fasta", "json"):#pragma: no cover
                self.services.logging.warning("frmt must be set to one of: txt, xml, rdf, gff, fasta, json.")
        elif database == 'uniparc':
            if frmt not in ( "xml", "rdf", "fasta", 'tsv', 'json'): #pragma: no cover
                raise ValueError("frmt must be set to one of: tsv, xml, rdf, gff, fasta, json")
                self.services.logging.warning("frmt must be set to one of: txt, xml, rdf, gff, fasta.")
        elif database == 'uniref':
            if frmt not in ("xml", "rdf", "fasta", 'tsv', 'json'): #pragma: no cover
                self.services.logging.warning("frmt must be set to one of: xml, rdf, gff, fasta, json.")
        elif database == "taxonomy":
            pass
        else: #pragma: no cover
            self.services.logging.warning("database must be set to uniref, uniparc, uniprot or taxonomy")


        if isinstance(uniprot_id, str):
            queries = uniprot_id.split(",")
        else:
            queries = uniprot_id
        #queries = self.services.devtools.to_list(uniprot_id)

        # some magic here not documented on uniprot website...but multiple queries are possible
        url = [database + "/" + query + "." + frmt for query in queries]

        # the frmt=txt here is for the requests, nothing related to the uniprot format
        res = self.services.http_get(url, frmt="txt", params={'include':include})
        if frmt == 'json':
            for i, x in enumerate(res):
                try:
                    res[i] = json.loads(x)
                except:
                    pass

        if isinstance(res, list) and len(res) == 1:
            res = res[0]
        return res

    def get_fasta(self, uniprot_id):
        """Returns FASTA string given a valid identifier

        :param str uniprot_id: a valid identifier (e.g. P12345)

        This is just an alias to :meth:`retrieve` when setting the format to 'fasta'. 
        Method kept for legacy.

        """
        res = self.retrieve(uniprot_id, frmt='fasta')
        return res

    def search(
        self,
        query,
        frmt="tsv",
        columns=None,
        include_isoforms=False,
        sort="score",
        compress=False,
        limit=None,
        offset=None,
        maxTrials=10,
        database="uniprotkb",
    ):
        """Provide some interface to the uniprot search interface.

        :param str query: query must be a valid uniprot query.
            See https://www.uniprot.org/help/query-fields and examples below
        :param str frmt: a valid format amongst html, tab, xls, asta, gff,
            txt, xml, rdf, list, rss. If tab or xls, you can also provide the
            columns argument.  (default is tab)
        :param str columns: comma-separated list of values. Works only if fomat
            is tsv or xls. For UnitProtKB, some possible columns are:
            id, entry name, length, organism. 
            See also :attr:`~bioservices.uniprot.UniProt.valid_mapping`
            for the full list of column keywords.
        :param bool include_isoform: include isoform sequences when the frmt
            parameter is fasta. Include description when frmt is rdf.
        :param str sort: by score by default. Set to None to bypass this behaviour
        :param bool compress: gzip the results
        :param int limit: Maximum number of results to retrieve.
        :param int offset:  Offset of the first result, typically used together
            with the limit parameter.
        :param int maxTrials: this request is unstable, so we may want to try
            several time.

        To obtain the list of uniprot ID returned by the search of zap70 can be
        retrieved as follows::

            >>> u.search('zap70+AND+organism_id:9606')
            >>> u.search("zap70+AND+taxonomy_id:9606", frmt="tsv", limit=3,
            ...    columns="entry_name,length,id, gene_names")
            Entry name  Length  Entry   Gene names
            CBLB_HUMAN  982 Q13191  CBLB RNF56 Nbla00127
            CBL_HUMAN   906 P22681  CBL CBL2 RNF55
            CD3Z_HUMAN  164 P20963  CD247 CD3Z T3Z TCRZ

        other examples::

            >>> u.search("ZAP70+AND+organism_id:9606", limit=3, columns="id,xref_pdb")

        You can also do a search on several keywords. This is especially useful
        if you have a list of known entry names.::

            >>> u.search("ZAP70_HUMAN+OR+CBL_HUMAN", frmt="tsv", limit=3,
            ...    columns="entry name,length,id, genes")
            Entry name  Length  Entry   Gene names


        Finally, note that when you search for a query, you may have several hits::

            >>> u.search("P12345)

        including the ID P12345 but also related entries. If you 
        need only the entry that perfectly match the query, use::

            >>> u.search("accession:P12345")

        This was provided from a user issue that was solved here:
        https://github.com/cokelaer/bioservices/issues/122


        .. warning:: some columns although valid may not return anything, not even in
            the header: 'score', 'taxonomy', 'tools'. this is a uniprot feature,
            not bioservices.

        .. versionchanged:: 1.10 
        
            Due to uniprot API changes in June 2022:

            * parameter 'include' is not named 'include_isoform
            * default parameter 'tab' is now 'tsv' but does not change the results

        """
        params = {}

        if frmt is not None:
            _valid_formats = [
                "xls",
                "fasta",
                "gff",
                "txt",
                "tsv",
                "xml",
                "rss",
                "list",
                "rss",
                "html",
            ]
            self.services.devtools.check_param_in_list(frmt, _valid_formats)
            params["format"] = frmt

        if columns is not None:
            self.services.devtools.check_param_in_list(frmt, ["tsv", "xls"])

            # remove unneeded spaces before/after commas if any
            if "," in columns:
                columns = [x.strip() for x in columns.split(",")]
            else:
                columns = [columns]


            # convert back to a string as expected by uniprot
            params["fields"] = ",".join([x.strip() for x in columns])

        if include_isoforms is True and frmt in ["fasta", "rdf"]:
            params["includeIsoform"] = "yes"

        if compress is True:
            params["compress"] = "yes"

        if sort:
            self.services.devtools.check_param_in_list(sort, ["score"])
            params["sort"] = sort

        if offset is not None:
            #if isinstance(offset, int):
            params["cursor"] = offset

        if limit is not None:
            if isinstance(limit, int):
                params["size"] = limit

        # + are interpreted and have a meaning. See arrayexpress module for details

        query = query.replace("+", " ")
        params['query'] = query
        del params['sort']

        res = self.services.http_get(f"{database}/search", frmt="txt", params=params)
        return res

    def quick_search(self, query, include_isoforms=False, sort="score", limit=None):
        """a specialised version of :meth:`search`

        This is equivalent to::

            u = uniprot.UniProt()
            u.search(query, frmt='tsv', sort="score", limit=None)

        :returns: a dictionary.

        """
        res = self.search(query, "tsv", include_isoforms=include_isoforms, 
            sort=sort, limit=limit)

        # if empty result, nothing to do
        if res and len(res) == 0:
            return res

        # else populate a dictionary
        newres = {}
        for line in res.split("\n")[1:-1]:
            # print line
            Entry, a, b, c, d, e, f = line.split("\t")
            # print Entry, a, b, c, d, e, f
            newres[Entry] = {
                "Entry name": a,
                "Status": b,
                "Protein names": c,
                "Gene names": d,
                "Organism": e,
                "Length": f,
            }
        return newres

    def uniref(self, query):
        """Calls UniRef service

        This is an alias to :meth:`retrieve`
        ::

            >>> u = UniProt()
            >>> u.uniref("Q03063")

        Another example from https://github.com/cokelaer/bioservices/issues/121
        is the combination of uniprot and uniref filters::

            u.uniref("uniprot:(ec:1.1.1.282 taxonomy_name:bacteria reviewed:true)")

        .. versionchanged:: 1.10 due to uniprot API changes in June 2022, 
            we now return a json instead of a pandas dataframe.
        """
        res = self.services.http_get(
            f"uniref/UniRef90_{query}.json", frmt="json"
        )
        return res

    def get_df(self, entries, nChunk=100, organism=None, limit=10):
        """Given a list of uniprot entries, this method returns a dataframe with all possible columns


        :param entries: list of valid entry name. if list is too large (about
            >200), you need to split the list
        :param chunk:
        :param limit: limit number of entries per identifier to 10. You can
            set it to None to keep all entries but this will be very slow
        :return: dataframe with indices being the uniprot id (e.g. DIG1_YEAST)

        To get about 100 columns related to the accession P62988, type:

            df = u.get_df('P62988')

        Note that you may preceed the accesion by the keyword **sec_acc) to access secondary 
        accessions numbers::

            df = u.get_df('sec_acc:P62988')

        """
        if isinstance(entries, str):
            entries = [entries]
        else:
            entries = list(set(entries))
        output = pd.DataFrame()

        self.services.logging.info(
            "fetching information from uniprot for {} entries".format(len(entries))
        )

        if limit < len(entries):
            self.services.logging.warning("The limit paramter is less than the number of entries. Probably not what you want. set limit to the length of the entries to remove this message")

        nChunk = min(nChunk, len(entries))
        N, rest = divmod(len(entries), nChunk)
        for i in range(0, N + 1):
            this_entries = entries[i * nChunk : (i + 1) * nChunk]
            if len(this_entries):
                self.services.logging.info("uniprot.get_df {}/{}".format(i + 1, N))
                query = "+OR+".join(this_entries)
                if organism:
                    query += f"+AND+{organism}"

                res = self.search(
                    query,
                    frmt="tsv",
                    columns=",".join(self._valid_columns),
                    limit=limit,
                )
            else:
                break

            if len(res) == 0:
                self.services.logging.warning("some entries %s not found" % entries)
            else:
                df = pd.read_csv(io.StringIO(str(res)), sep="\t")

                if isinstance(output, type(None)):
                    output = df.copy()
                else:
                    # output = output.append(df, ignore_index=True)
                    output = pd.concat([output, df], ignore_index=True)

        # you may end up with duplicated...
        output.drop_duplicates(inplace=True)
        # Sequences are splitted into chunks of 10 characters. let us remove
        # the spaces:
        if "sequence" in output.columns:
            output["sequence"].fillna("", inplace=True)
            output.Sequence = output["sequence"].apply(lambda x: x.replace(" ", ""))

        return output



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2015 - EBI-EMBL
#  Copyright (c) 2016-2017 - Institut Pasteur
#
#  File author(s): thomas cokelaer
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`~Seqret` to access to Seqret WS.


.. topic:: What is Seqret ?

    :URL: http://www.ebi.ac.uk/Tools/services/rest/seqret/
    :Service:
    :Citations: http://www.ncbi.nlm.nih.gov/pubmed/18428689

    .. highlights::

        EMBOSS seqret reads and converts biosequences between a selection of common
        biological sequence formats, including EMBL, GenBank and fasta sequence
        formats.

        Seqret homepage -- Sep 2017


"""
import copy
from bioservices.services import REST


__all__ = ["Seqret"]


class Seqret:
    """Interface to the `Seqret <http://www.ebi.ac.uk/readseq>`_ service

    ::

        >>> from bioservices import *
        >>> s = Seqret()

    The ReadSeq service was replaced by #the Seqret services (2015).

    .. versionchanged:: 0.15

    """

    def __init__(self, verbose=True):
        """.. rubric:: Constructor

        :param bool verbose:

        """
        url = "https://www.ebi.ac.uk/Tools/services/rest/emboss_seqret"
        self.services = REST(name="seqret", url=url, verbose=verbose)
        self._parameters = None

    def get_parameters(self):
        """Get a list of the parameter names.

        :returns: a list of strings giving the names of the parameters.

        """
        parameters = self.services.http_get("parameters", frmt="json")

        return parameters["parameters"]

    def _get_parameters(self):
        if self._parameters:
            return self._parameters
        else:
            res = self.get_parameters()
            self._parameters = res
        return self._parameters

    parameters = property(_get_parameters, doc="Get list of parameter names")

    def get_parameter_details(self, parameterId):
        """Get details of a specific parameter.

        :param str parameter: identifier/name of the parameter to fetch details of.
        :return: a data structure describing the parameter and its values.

        ::

            rs = ReadSeq()
            print(rs.get_parameter_details("stype"))

        """
        if parameterId not in self.parameters:
            raise ValueError("Invalid parameterId provided(%s). See parameters attribute" % parameterId)

        request = "parameterdetails/" + parameterId
        res = self.services.http_get(request, frmt="json")
        return res

    def run(self, email, title, **kargs):
        """Submit a job to the service.

        :param str email: user e-mail address.
        :param str title: job title.
        :param params: parameters for the tool as returned by :meth:`get_parameter_details`.
        :return: string containing the job identifier (jobId).

        Deprecated (olf readseq service)::

            Format Name     Value
            Auto-detected   0
            EMBL            4
            GenBank         2
            Fasta(Pearson)  8
            Clustal/ALN     22
            ACEDB           25
            BLAST           20
            DNAStrider      6
            FlatFeat/FFF    23
            GCG             5
            GFF             24
            IG/Stanford     1
            MSF             15
            NBRF            3
            PAUP/NEXUS      17
            Phylip(Phylip4)     12
            Phylip3.2       11
            PIR/CODATA      14
            Plain/Raw       13
            SCF             21
            XML             19

        As output, you also have

        Pretty 18

        ::

            s = readseq.Seqret()
            jobid = s.run("cokelaer@test.co.uk", "test", sequence=fasta, inputformat=8,
                outputformat=2)
            genbank = s.get_result(s._jobid)


        """
        for k in kargs.keys():
            self.services.devtools.check_param_in_list(k, self.parameters)

        assert "sequence" in kargs.keys()
        params = {"email": email, "title": title}

        for k in [
            "stype",
            "inputformat",
            "outputformat",
            "feature",
            "firstonly",
            "reverse",
            "outputcase",
            "seqrange",
        ]:
            if k in kargs.keys():
                value = kargs.get(k)
                details = self.get_parameter_details(k)
                valid_values = [x["value"] for x in details["values"]["values"]]
                self.services.devtools.check_param_in_list(str(value), valid_values)
                params[k] = value
        # r = requests.post(url + "/run?", data={"sequence":fasta, "stype": "protein",
        # "inputformat":"raw", "outputformat":"fasta", "email":"thomas.cokelaer@pasteur.fr",
        # "title":"test"})

        params["sequence"] = kargs["sequence"]

        jobid = self.services.http_post("run", frmt="txt", data=params)
        self._jobid = jobid
        return jobid

    def get_status(self, jobid=None):
        """Get the status of a submitted job.

        :param str jobid: job identifier.
        :return: string containing the status.

        The values for the status are:

        - RUNNING: the job is currently being processed.
        - FINISHED: job has finished, and the results can then be retrieved.
        - ERROR: an error occurred attempting to get the job status.
        - FAILURE: the job failed.
        - NOT_FOUND: the job cannot be found.

        """
        res = self.services.http_get("status/{}".format(jobid), frmt="txt")
        return res

    def get_result_types(self, jobid):
        """Get the available result types for a finished job.

        :param str jobid: job identifier.
        :return: a list of wsResultType data structures describing the available result types.
        """
        res = self.services.http_get("resulttypes/{}".format(jobid), frmt="json")
        return [x["identifier"] for x in res["types"]]

    def get_result(self, jobid, result_type="out"):
        """Get the result of a job of the specified type.

        :param str jobid: job identifier.
        :param parameters: optional list of wsRawOutputParameter used to
            provide additional parameters for derived result types.
        """
        if self.get_status(jobid) != "FINISHED":
            self.services.logging.warning("Your job is not finished yet. Try again later.")
            return

        # result_types = self.get_result_types(jobid)
        # assert parameters in result_types
        res = self.services.http_get("result/{}/{}".format(jobid, result_type), frmt="txt")

        return res



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EMBL-EBI
#
#  File author(s):
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module includes common tools to manipulate XML files"""
import xml.etree.ElementTree as ET
import bs4

try:
    from urllib.request import urlopen
    from urllib.error import HTTPError
except:
    from urllib2 import urlopen, HTTPError

__all__ = ["easyXML", "readXML"]


class easyXML(object):
    """class to ease the introspection of XML documents.

    This class uses the standard xml module as well as the package BeautifulSoup
    to help introspecting the XML documents.

    ::

        >>> from bioservices import *
        >>> n = ncbiblast.NCBIblast()
        >>> res = n.getParameters() # res is an instance of easyXML
        >>> # You can retreive XML from this instance of easyXML and print the content
        >>> # in a more human-readable way.
        >>> res.soup.findAll('id') # a Beautifulsoup instance is available
        >>> res.root # and the root using xml.etree.ElementTree

    There is a getitem so you can type::

        res['id']

    which is equivalent to::

        res.soup.findAll('id')

    There is also aliases findAll and prettify.

    """

    def __init__(self, data, encoding="utf-8"):
        """.. rubric:: Constructor

        :param data: an XML document format
        :param fixing_unicode: use only with HGNC service to fix issue with the
            XML returned by that particular service. No need to use otherwise.
            See :class:`~bioservices.hgnc.HGNC` documentation for details.
        :param encoding: default is utf-8 used. Used to fix the HGNC XML only.


        The data parameter must be a string containing the XML document. If you
        have an URL instead, use :class:`readXML`

        """
        # if fixing_unicode:
        #    x = unicodefix.FixingUnicode(data, verbose=False, encoding=encoding)
        #    self.data = x.fixed_string.encode("utf-8")
        # else:
        self.data = data[:]

        try:
            self.root = ET.fromstring(self.data)
        except:
            self.root = self.data[:]
        self._soup = None
        self.prettify = self.soup.prettify
        self.findAll = self.soup.findAll

    def getchildren(self):
        """returns all children of the root XML document

        This is just an alias to self.soup.getchildren()
        """
        return [x for x in self.root]

    def _get_soup(self):
        if self._soup is None:
            self._soup = bs4.BeautifulSoup(self.data, "lxml")
        return self._soup

    soup = property(_get_soup, doc="Returns the beautiful soup instance")

    def __str__(self):
        txt = self.soup.prettify()
        return txt

    def __getitem__(self, i):
        return self.findAll(i)


class readXML(easyXML):
    """Read XML and converts to beautifulsoup data structure

    easyXML accepts as input a string. This class accepts a filename instead
    inherits from easyXML

    .. seealso:: :class:`easyXML`

    """

    def __init__(self, url, encoding="utf-8"):
        self.data = urlopen(url).read()
        super(readXML, self).__init__(self.data, encoding)


class XMLObjectify(object):
    def __init__(self, obj):
        """obj can be easyXML data set"""
        from lxml import objectify

        try:
            self.root = objectify.fromstring(obj.data)
        except:
            # try something else
            self.root = objectify.fromstring(obj)
        self.obj = obj

    def __str__(self):
        txt = ""
        for child in self.root.getchildren():
            txt += child.tag + "\n"
        return txt



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to DBFetch web service

.. topic:: What is DBFetch

    :URL: http://www.ebi.ac.uk/Tools/webservices/services/dbfetch
    :Service: http://www.ebi.ac.uk/Tools/webservices/services/dbfetch_rest

    .. highlights::

        "DBFetch allows you to retrieve entries from various up-to-date biological
        databases using entry identifiers or accession numbers. This is equivalent to
        the CGI based dbfetch service. Like the CGI service a request can return a
        maximum of 200 entries."

        -- From http://www.ebi.ac.uk/Tools/webservices/services/dbfetch , Dec 2012


"""
from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["DBFetch"]


class DBFetch(REST):
    """Interface to `DBFetch <http://www.ebi.ac.uk/Tools/webservices/services/dbfetch_rest>`_ service

    ::

        >>> from bioservices import DBFetch
        >>> w = DBFetch()
        >>> data = w.fetchBatch("uniprot" ,"zap70_human", "xml", "raw")

    For more information about the API, check this page:
    http://www.ebi.ac.uk/Tools/dbfetch/syntax.jsp

    """

    _url = "http://www.ebi.ac.uk/Tools/dbfetch"

    def __init__(self, verbose=False):
        """.. rubric:: Constructor

        :param bool verbose: print informative messages
        """
        super(DBFetch, self).__init__(name="DBfetch", url=DBFetch._url, verbose=verbose)
        self._supportedDBs = None
        self._supportedFormats = None
        self._supportedStyles = None

    def _check_db(self, db):
        if db not in self.supported_databases:
            raise Exception("%s not a supportedDB. " % db)

    def fetch(self, query, db="ena_sequence", format="default", style="raw", pageHtml=False):
        """Fetch an entry in a defined format and style.

        :param str query: the entry identifier in db:id format (e.g. 'UniProtKB:WAP_RAT').
        :param str format: the name of the format required (default to fasta).
        :param str style: the name of the style required (raw, default, html)

        :returns: The format of the response depends on the format/style
            parameter.


        ::

            from bioservices import DBFetch
            u = DBFfetch()
            db.fetch(db="ena_sequence", format="fasta", query="L12344,L12345")
            db.fetch(db="uniprot", format="fasta", query="P53503")


        If db is ommited, the default is ena_sequence.
        If formatare ommited, the default is EMBL format
        The default style is raw data.

        """
        self._check_db(db)
        res = self.http_get(
            "dbfetch",
            params={
                "db": db,
                "format": format,
                "style": style,
                "id": query,
                "pageHtml": pageHtml,
            },
        )
        try:
            res = res.content.decode()
        except:
            pass
        return res

    def get_database_info(self, db=None):
        """Get details describing specific database (data formats, styles)

        :param str db: a valid database.
        :return: The output can be introspected and contains several attributes

        ::

            >>> res = u.get_database_info('uniprotkb')
            >>> print(res['description'])
            'The UniProt Knowledgebase (UniProtKB) is the central access point for extensive curated protein information, including function, classification, and cross-references. Search UniProtKB to retrieve everything that is known about a particular sequence.'

        """
        res = self.http_get("dbfetch/dbfetch.databases?style=json")
        if db:
            self._check_db(db)
            res = res[db]
        return res

    def get_all_database_info(self):
        """Get details of all available databases, includes formats and result styles.

        :Returns: A list of data structures describing the databases. See
            :meth:`getDatabaseInfo` for a description of the data structure.
        """
        return self.get_database_info()

    def get_database_formats(self, db):
        """Get list of format names for a given database.

        :param str db: valid database name

        ::

            >>> db.get_database_formats("uniprotkb")
            ['default',
             'annot',
             'entrysize',
             'fasta',
             'gff3',
             'seqxml',
             'uniprot',
             'uniprotrdfxml',
             'uniprotxml',
             'dasgff',
             'gff2']

        """
        self._check_db(db)
        res = self.http_get("dbfetch?info=formats&db={}".format(db)).content
        res = res.decode().split()
        return res

    def get_database_format_styles(self, db, format):
        """Get a list of style names available for a given database and format.

        :param str db: database name to get available styles for (e.g. uniprotkb).
        :param str format: the data format to get available styles for (e.g. fasta).

        :Returns: An array of strings containing the style names.

        ::

            >>> u.get_database_format_styles("uniprotkb", "fasta")
            ['default', 'raw', 'html']

        """
        self._check_db(db)
        res = self.http_get("dbfetch?info=styles&format={}&db={}".format(format, db)).content
        res = res.decode().split()
        return res

    def _getSupportedDBs(self):
        """Get a list of database names usable with DBFetch.

        Buffered in _supportedDB.
        """
        if self._supportedDBs:
            return self._supportedDBs
        else:
            res = self.http_get("dbfetch?info=dbs").content
            self._supportedDBs = res.decode().split() + ["default"]
        return self._supportedDBs

    supported_databases = property(_getSupportedDBs, doc="Alias to getSupportedDBs.")



#
#  This file is part of BioServices
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/sequana/sequana
#  documentation: http://sequana.readthedocs.io
#
##############################################################################
import functools
import glob
import os
from pathlib import Path
import pkg_resources
import shutil
import subprocess
import sys
import tempfile

import click

import colorlog

logger = colorlog.getLogger(__name__)

__all__ = ["main"]


# This can be used by all commands as a simple decorator
def common_logger(func):
    @click.option(
        "--logger",
        default="INFO",
        type=click.Choice(["INFO", "DEBUG", "WARNING", "CRITICAL", "ERROR"]),
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


from bioservices import version


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version)
def main(**kwargs):
    """This is the main entry point for a set of BioServices application"""
    pass


@main.command()
@click.option("--accession", type=click.STRING, required=True,
    help="A valid accession number (e.g., FN433596.1)")
@click.option("--prefix", type=click.STRING, default=None,
    help="""By default, output FastA file is name after the accession number
            but you can rename it using this --prefix name""")
@click.option("--method", type=click.Choice(["ENA", "EUtils"]), default="EUtils")
@click.option("--with-gff3", is_flag=True, default=False)
@click.option("--with-gbk", is_flag=True, default=False)
def download_accession(**kwargs):
    """Download Fasta related to an accession and possibly other type (e.g gff)

    Input file can be gzipped or not. The --output-file

        bioservices download-accession FN433596.1
    """
    from bioservices.apps.download_fasta import download_fasta
    from bioservices.apps.download_gff3 import download_gff3
    from bioservices.apps.download_gbk import download_gbk

    prefix = kwargs['prefix']

    logger.info("Downloading FastA file")
    download_fasta(kwargs["accession"], output_filename=f"{prefix}.fa" if prefix else prefix, 
        method=kwargs["method"])

    if kwargs["with_gff3"]:
        logger.info("Downloading GFF file")
        download_gff3(kwargs["accession"], output_filename=f"{prefix}.gff" if prefix else prefix, 
            method=kwargs["method"])

    if kwargs["with_gbk"]:
        logger.info("Downloading Genbank file")
        download_gbk(kwargs["accession"], output_filename=f"{prefix}.gbk" if prefix else prefix, 
            method=kwargs["method"])






#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      https://github.com/cokelaer/bioservices
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  source: http://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to OmniPath web service

.. topic:: What is OmniPath ?

    :URL: http://omnipathdb.org
    :URL: https://github.com/saezlab/pypath/blob/master/webservice.rst

    .. highlights::

        A comprehensive collection of literature curated human signaling pathways.

        -- From OmniPath web site, March 2016


"""
from bioservices import REST
from easydev import to_list
from bioservices import logger

logger.name = __name__


class OmniPath(REST):
    """Interface to the `OmniPath <http://www.ebi.ac.uk/unichem/>`_ service

    .. doctest::

            >>> from bioservices import OmniPath
            >>> o = OmniPath()
            >>> net = o.get_network()
            >>> interactions = o.get_interactions('P00533')

    """

    _url = "http://omnipathdb.org/"

    def __init__(self, verbose=False, cache=False):
        """**Constructor** OmniPath

        :param verbose: set to False to prevent informative messages
        """
        super(OmniPath, self).__init__(name="OmniPath", url=OmniPath._url, verbose=verbose, cache=cache)

    def get_about(self):
        """Information about the version"""
        res = self.http_get(self.url + "about").content
        return res

    def get_network(self, frmt="json"):
        """Get basic statistics about the whole network including sources"""
        assert frmt in ["json", "tsv"], "frmt must be set to json or tsv"
        res = self.http_get(self.url + "network", frmt=frmt, params={"format": frmt})

        return res

    def get_interactions(self, query="", frmt="json", fields=[]):
        """Interactions of proteins

        :param str query: a valid uniprot identifier (e.g. P00533). It can also
            be a list of uniprot identifiers, or a string with
            comma-separated identifiers.
        :param str fields: additional fields to be added to the output
            (e.g., sources, references)
        :param str frmt: format of the output (json or tabular)


        Example::

            res_one = o.get_interactions('P00533')
            res_many = o.get_interactions('P00533,O15117,Q96FE5')
            res_many = o.get_interactions(['P00533','O15117','Q96FE5'])


            res_one = o.get_interactions('P00533', fields='sources')
            res_one = o.get_interactions('P00533', fields=['source'])
            res_one = o.get_interactions('P00533', fields=['source', 'references'])

        You may also keep query to an empty string, but the entire DB will then
        be downloaded. This may take time and the :attr:`timeout` may need to be
        increased manually.

        If frmt is set to TSV, the output is a TSV table with a header. If set
        to json, a dictionary is returned.
        """
        # make sure there is no spaces
        if isinstance(query, list):
            query = ",".join(query)
        else:
            try:  # if input is a string
                query = query.replace(" ", "")
            except:
                pass
        assert frmt in ["json", "tsv"], "frmt must be set to json or tsv"
        params = {}
        params["format"] = frmt
        from easydev import to_list

        fields = to_list(fields)

        if len(fields):
            params["fields"] = fields

        # TODO handle multiple fields
        res = self.http_get(self.url + "interactions/%s" % query, frmt=frmt, params=params)
        return res

    def get_resources(self, frmt="json"):
        """Return statistics about the databases and their contents"""
        res = self.http_get(self.url + "resources", frmt=frmt, params={"format": frmt})
        return res

    def get_info(self):
        """Currently returns HTML page"""
        from easydev import browser

        browser.browse(self.url + "info")

    def get_ptms(self, query="", ptm_type=None, frmt="json", fields=[]):
        """List enzymes, substrates and PTMs

        :param str query: a valid uniprot identifier (e.g. P00533). It can also
            be a list of uniprot identifiers, or a string with
            comma-separated identifiers.
        :param str ptm_type: restrict the output to this type of PTM
            (e.g., phosphorylation)
        :param str fields: additional fields to be added to the output
            (e.g., sources, references)


        """
        # make sure there is no spaces
        if isinstance(query, list):
            query = ",".join(query)
        else:
            try:  # if input is a string
                query = query.replace(" ", "")
            except:
                pass
        assert frmt in ["json", "tsv"], "frmt must be set to json or tsv"
        params = {}
        params["format"] = frmt
        if len(fields):
            params["fields"] = fields

        res = self.http_get(self.url + "ptms/%s" % query, frmt="json", params=params)
        return res



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the Pfam web service

.. topic:: What is Pfam ?

    :URL: http://www.uniprot.org
    :Citation:

    .. highlights::


        -- From Pfam  web site (help/about), Aug 2013



"""
from bioservices.services import REST
from bioservices import logger

logger.name = __name__

__all__ = ["Pfam"]


class Pfam(REST):
    """Interface to `Pfam <http://pfam.sanger.ac.uk>`_ pages

    This is not a REST interface actually but rather a parser to some of the
    HTML pages relatd to pathways.

    One can retrieve the pathways names and their list of proteins.

        >>> from bioservics import *
        >>> p = Pfam()

    """

    _url = "http://pfam.xfam.org/"

    def __init__(self, verbose=True):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        super(Pfam, self).__init__(name="Pfam", url=Pfam._url, verbose=verbose)

    def show(self, Id):
        """Just an example of opening a web page with a uniprot Id

        p = Pfam()
        p.show("P43403")

        """
        url = self._url + "/protein/" + Id
        self.on_web(url)

    def get_protein(self, ID):
        res = self.http_get("protein", params={"id": ID, "output": "xml"})
        return res.content



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`ChEMBL`

.. topic:: What is ChEMBL

    :URL:  https://www.ebi.ac.uk/chembl
    :REST: https://www.ebi.ac.uk/chembl/api/data

    .. highlights::

        "Using the ChEMBL web service API users can retrieve data from the ChEMBL
        database in a programmatic fashion. The following list defines the currently
        supported functionality and defines the expected inputs and outputs of each
        method."

        -- From ChEMBL web page Dec 2012

"""
import os
from bioservices.services import REST
from tqdm import tqdm
import webbrowser
from bioservices import logger

logger.name == __name__
try:
    from urllib.parse import quote
except:
    from urllib import quote

__all__ = ["ChEMBL"]


class ChEMBL:
    """New ChEMBL API bioservices 1.6.0

    **Resources**

    ChEMBL database is made of a set of resources. We recommend to look at
    https://arxiv.org/pdf/1607.00378.pdf

    Here we first create an instance and retrieve the first 1000 molecules from
    the database using the **limit** parameter.

    .. doctest::

        >>> from bioservices import ChEMBL
        >>> c = ChEMBL()
        >>> res = c.get_molecule(limit=1000)

    The returned objet is a list of 1000 records, each of them being a
    dictionary. The **molecule** resource is actually a very large one
    and one may want to skip some entries. This is possible using the **offset**
    parameter as follows::

        # Retrieve 1000 molecules skipping the first 50
        res = c.get_molecule(limit=1000, offset=50)

    If you want to know all resources available and the number of entries in
    each resources, use::

        status = c.get_status_resources()

    For instance, you should be able to get the total number of entries in the
    *mechanism* resource is about 5,000::

        print(status['mechanism'])

    To retrieve all entries from the mechanism resource, you can either set
    limit to a value large enough::

        res = c.get_mechanism(limit=1000000)

    or simply set it to -1::

        res = c.get_mechanism(limit=-1)

    All resources methods behaves in the same way.

    Those resources methods are: :meth:`get_activity`, :meth:`get_assay`,
    :meth:`get_atc_class`, :meth:`get_binding_site`, :meth:`get_biotherapeutic`,
    :meth:`get_cell_line`, :meth:`get_chembl_id_lookup`,
    :meth:`get_compound_record`, :meth:`get_compound_structural_alert`,
    :meth:`get_document`, :meth:`get_document_similarity`,
    :meth:`get_document_term`, :meth:`get_drug`, :meth:`get_drug_indication`,
    :meth:`get_go_slim`, :meth:`get_mechanism`, :meth:`get_metabolism`,
    :meth:`get_molecule`, :meth:`get_molecule_form`, :meth:`get_protein_class`,
    :meth:`get_source`, :meth:`get_target`, :meth:`get_target_component`,
    :meth:`get_target_prediction`, :meth:`get_target_relation`,
    :meth:`get_tissue`.

    **3 ways of getting items**

    1. Retrieve everything::

        c.get_molecule(limit=-1)

    2. Retrieve a specific entry::

        c.get_molecule("CHEMBL24")

    3. Retrieve a set of entries::

        c.get_molecule(["CHEMBL24","CHEMBL2"])

    **Filtering and Ordering**

    For ordering the results, we provide a simple method :meth:`order_by` that
    allows to sort the dictionary according to values in a specific key.

    Any data returned by a resource method (a list of dictionary) can be process
    through this method::

        c = ChEMBL()
        data = c.get_drug(limit=100)
        ordered_data = c.order_by(data, 'chirality')

    If you want to order using a key within a key, for instance order by
    molecular weight stored in the *molecular_properties* key, use the double
    underscore method as follows::

        c = ChEMBL()
        data = c.get_drug(limit=100)
        ordered_data = c.order_by(data, 'molecular_properties__mw_freebase')

    For filtering, it is possible to apply search filters to any resources.
    For example, it is possible to return all ChEMBL targets that contain
    the term 'kinase' in the pref_name attribute::

        c.get_target(filters='pref_name__contains=kinase")

    The pattern for applying a filter is as follows::

        [field]__[filter_type]=[value]

    where field has to be found by the user. Simply introspect the content of an
    item returned by the resource. For instance::

        c.get_target(limit=1) # to get one entry

    Let us consider the case of the **molecule** resource. You can
    retrieve the first 10 molecules using e.g.::

        res = c.get_molecule(limit=10)

    If you look at the first entry using res[0], you will get about
    38 keys. For instance **molecule_properties** or
    **molecule_chembl_id**.

    You can filter the molecules to keep only the molecule_chembl_id
    that match either CHEMBL25 or CHEMBL1000 using::

        res = c.get_molecule(filters='molecule_chembl_id__in=CHEMBL25,CHEMBL1000')

    For **molecule_properties**, this is actually a dictionary. For instance,
    inside the **molecule_properties** field, you have the molecular weight
    (mw_freebase). So to apply this filter, you need to use the following code
    (to keep molecules with molecular weight greater than 300::

        res = c.get_molecule(filters='molecule_properties__mw_freebase__gte=300')

    Here are the different types of filtering:

    =============== =============================================
    Filter Type     Description
    =============== =============================================
    exact (iexact)  Exact match with query
    contains        wild card search with query
    startswith      starts with query
    endswith        ends with query
    regex           regulqr expression query
    gt (gte)        Greater than (or equal)
    lt (lte)        Less than (or equal)
    range           Within a range of values
    in              Appears within list of query values
    isnull          Field is null
    search          Special type of filter allowing a full text
                    search based on Solr queries.
    =============== =============================================

    Several filters can be applied at the same time using a list::

        filters = ['molecule_properties__mw_freebase__gte=300']
        filters += ['molecule_properties__alogp__gte=3']
        res = c.get_molecule(filters)

    **Use Cases: (inspired from ChEMBL documentation)**

    Search molecules by synonym::

        >>> from bioservices import ChEMBL
        >>> c = ChEMBL()
        >>> res = c.search_molecule('aspirin')

    or SMILE, or InChiKey, or CHEMBLID::

        >>> res = c.get_molecule("CC(=O)Oc1ccccc1C(=O)O")
        >>> res = c.get_molecule("BSYNRYMUTXBXSQ-UHFFFAOYSA-N")
        >>> res = c.get_molecule('CHEMBL25')

    Several molecules at the same time can also be retrieved using lists::

        >>> res = c.get_molecule(['CHEMBL25', 'CHEMBL2'])



    Search target by gene name::

        >>> res = c.search_target("GABRB2")
        >>> len(res['targets'])
        18

    or directly in the target synonym field::

        >>> res = c.get_target(filters='target_synonym__icontains=GABRB2')

    .. note:: Not sure what is the difference between icontains vs contains.
        It looks like icontains is more permissive (you get more entries
        with icontains).

    Having a list of molecules ChEMBL IDs in a list, get uniprot accession
    numbers that map to those compounds::

        # First, get some IDs of approved drugs (about 2000 molecules)
        c = ChEMBL()
        drugs = c.get_approved_drugs()
        IDs = [x['molecule_chembl_id'] for x in drugs]

        # we jump from compounds to targets through activities
        # Here this is a one to many mapping so we initialise a default
        # dictionary.
        compound2target = defaultdict(set)

        filter = "molecule_chembl_id__in={}"
        for i in range(0, len(IDs), 50):
            activities = c.get_activity(filter.format(IDs[i:i+50]))
            # get target ChEMBL IDs from activities
            for act in activities:
                compound2target[act['molecule_chembl_id']].add(act['target_chembl_id'])

        # What we need is to get targets for all targets found in the previous
        # step. For each compound/drug there are hundreds of targets though. And
        # we will call the get_target for each list of hundreds targets. This
        # will take forever. Instead, because there are *only* 12,000 targets,
        # let us download all of them ! This took about 4 minutes on this test but
        # if you use the cache, next time it will be much much quicker. This is
        # not down at the activities level because there are too many entries

        targets = c.get_target(limit=-1)

        # identifies all target chembl id to easily retrieve the entry later on
        target_names = [target['target_chembl_id'] for target in targets]

        # retrieve all uniprot accessions for all targets of each compound
        for compound, targs in compounds2targets.items():
            accessions = set()
            for target in targs:
                index = target_names.index(target)
                accessions = accessions.union([comp['accession']
                    for comp in targets[index]['target_components']])
            compounds2targets[compound] = accessions

    In version 1.6.0 of bioservices, you can simply use::

        res = c.compounds2targets(IDs)


    Get Target type count for all targets::

        import collections
        collections.Counter([x['target_type'] for x in targets]


    Find compounds similar to given SMILES query with similarity threshold of
    85%::

        >>> SMILE = "CN(CCCN)c1cccc2ccccc12"
        >>> c.get_similarity(SMILE, similarity=70)

    Find compounds similar to aspirin (CHEMBL25) with similarity
    threshold of 70%::

        # search for aspirin in all molecules and from first hist
        # get the ChEMBL ID
        >>> molecules = c.search_molecule("aspirin")['molecules']
        >>> chembl_id = molecules[0]['molecule_chembl_id']
        # now use the :meth:`get_similarity` given the ID
        >>> res = c.get_similarity(chembl_id, similarity=70)

    Perform substructure search using SMILES or ChEMBID::

        >>> res = c.get_substructure("CN(CCCN)c1cccc2ccccc12")
        >>> res = c.get_substructure("CHEMBL25")


    Obtain he pChEMBL value for compound::

        res = c.get_activity(filters=['pchembl_value__isnull=False',
                                      'molecule_chembl_id=CHEMBL25'])


    Obtain he pChEMBL value for compound and target::

        res = c.get_activity(filters=['pchembl_value__isnull=False',
                                      'molecule_chembl_id=CHEMBL25',
                                      'target_chembl_id=CHEMBL612545'])

    Get all approved drugs::

        c.get_approved_drugs(max_phase=4)

    Get approved drugs for lung cancer


    The ChEMBL API significantly changed in 2018 and the new version of
    bioservices (1.6.0) had to change the API as well, which has been
    simplified.

    Here below are some correspondances between the previous and the new API.

    ========================================== ==========================
    bioservices before 1.6.0                   After 1.6.0
    ========================================== ==========================
    get_compounds_substructure                 get_substructure
    get_compounds_similar_to_SMILES            get_similarity(SMILE)
    get_compounds_by_chemblId(ID)              get_similarity(ID)
    get_individual_compounds_by_inChiKey       get_molecule(inchikey)
    get_compounds_by_chemblId_form             get_molecule_form
    get_compounds_by_chemblId_drug_mechanism   get_mechanism(ID)
    get_target_by_chemblId(ID)                 get_target(ID)
    get_image_of_compounds_by_chemblId         get_image
    etc
    ========================================== ==========================

    :references:
        - https://arxiv.org/pdf/1607.00378.pdf
        - https://www.ebi.ac.uk/chembl/api/data/docs
    """

    _url = "https://www.ebi.ac.uk/chembl/api/data"

    def __init__(self, verbose=False, cache=False):
        self.services = REST(name="ChEMBL", url=ChEMBL._url, verbose=verbose, cache=cache, 
            url_defined_later=True)
        self.format = "json"

    def _get_data(self, name, params):

        # keep the number of events we want and original offset
        max_data = params["limit"]
        offset = params["offset"]

        # I noticed that
        # if offset + limit > total_count, then limit is set to 1000 - offset
        # Not sure whether it is a bug or intended behaviour but this caused
        # some issues during the debugging.

        # So services.http_get("mechanism?format=json&limit=10000&offset=10")
        # returns 990 entries and not 1000 as expected.

        # if a resources is small (e.g. tissue has 655 < 1000 entries) there is
        # no such issues.

        # So, the best is to constraint limit to 1000
        params["limit"] = 1000  # for the first call

        # The limit used in all other calls
        limit = 1000

        res = self.services.http_get("{}".format(name), params=params)
        self._check_request(res)

        # get rid of page_meta key/value
        self.page_meta = res["page_meta"]
        keys = list(res.keys())
        keys.remove("page_meta")
        names = keys[0]  # the parameter name in plural form

        # keep first chunk of data
        data = res[names]

        if max_data == -1:
            max_data = res["page_meta"]["total_count"]
        elif max_data > res["page_meta"]["total_count"]:
            max_data = res["page_meta"]["total_count"]

        N = max_data

        count = 1

        with tqdm(total=N, desc='TEST') as pb:
            while res["page_meta"]["next"] and len(data) < max_data:
                params["limit"] = limit
                params["offset"] = limit * count + offset
                res = self.services.http_get("{}".format(name), params=params)
                data += res[names]
                count += 1
                pb.update(len(data))
                self.page_meta = res["page_meta"]

        if self.page_meta["next"]:
            offset = self.page_meta["offset"]
            total = self.page_meta["total_count"] - len(data) - int(offset)
            self.services.logging.warning(
                "More data available ({}). rerun with higher"
                "limit and/or offset {}. Check content of page_meta"
                " attribute".format(total, offset)
            )

        if len(data) > max_data:
            return data[0:max_data]
        else:
            return data

    def _check_request(self, res):
        # If there is no output because of wrong query, a 404 is returned.
        if isinstance(res, int):
            raise ValueError("Invalid request for {} {}. Check your query and parameters")

    def _get_this_service(self, name, query, params={"limit": 20, "offset": 0}):
        """


        if query is None, calls the resources

            URL/data/[resource]

        if query is a string, calls

            URL/data/[resource]/ID

        if query is a list of IDS, calls

            URL/data/[resource]/set/[IDS]

        In case 1 and 3, returns a dictionary and populate attribute page_meta.
        In case 2, there is only one requested ID so returns a dictionary (not a
        list of dictionaries).

        """

        # look at any filters provided by the user
        if params["filters"] is None:
            del params["filters"]
        elif isinstance(params["filters"], list):
            for filter in params["filters"]:
                assert filter.count("=") == 1
                key, value = filter.split("=")
                params[key] = value
            del params["filters"]
        else:
            assert params["filters"].count("=") == 1
            k, v = params["filters"].split("=")
            del params["filters"]
            params[k] = v

        params["format"] = self.format
        # Here, we will switch between several ways of using each
        # service.
        if query is None:
            res = self._get_data(name, params)
        # user may use integer, floats or strings.
        elif isinstance(query, (str, int, float)):
            res = self.services.http_get("{}/{}".format(name, query), params=params)
            self._check_request(res)
        elif isinstance(query, list):
            assert params["limit"] <= 1000, "limit must be less than 1000"
            ids = ";".join([str(x) for x in query])
            res = self.services.http_get("{}/set/{}".format(name, ids), params=params)
            self._check_request(res)
            # Note that there is no page_meta key in the returned object but a
            # single key that is the plural for of the resource except if some
            # entries are not found. In such case, a
            if "not_found" in res.keys():
                self.services.logging.warning("Some entries were not found: {}".format(res["not_found"]))
                self.not_found = res["not_found"]
                del res["not_found"]
            names = list(res.keys())[0]
            res = res[names]

        return res

    def _search(self, name, query, params):
        # Check the validity of limits
        assert params["limit"] > 0, "limits must be less than 1000"
        assert params["limit"] <= 1000, "limits must be positive"
        res = self.services.http_get("{}/search.{}?q={}".format(name, self.format, query), params=params)

        if isinstance(res, int):
            self.services.logging.warning("Invalid request for {} {}. Check your parameters".format(name, params))
            return {}

        if "page_meta" in res and res["page_meta"]["next"]:
            Next = res["page_meta"]["next"]
            offset = Next.split("&offset=")[1]
            self.services.logging.warning("More data available with offset {}".format(offset))
        return res

    def search_activity(self, query, limit=20, offset=0):
        """Activity values recorded in an Assay"""
        params = {"limit": limit, "offset": offset}
        return self._search("activity", query, params=params)

    def get_activity(self, query=None, limit=20, offset=0, filters=None):
        """Activity values recorded in an Assay"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("activity", query, params=params)

    def search_assay(self, query, limit=20, offset=0):
        """Assay details as reported in source document"""
        params = {"limit": limit, "offset": offset}
        return self._search("assay", query, params=params)

    def get_assay(self, query=None, limit=20, offset=0, filters=None):
        """Assay details as reported in source Document/Dataset


        >>> c.get_assay("CHEMBL1217643")

        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("assay", query, params=params)

    def get_ATC(self, limit=20, offset=0, filters=None):
        """WHO ATC Classification for drugs

        c.get_atc()
        c['atc']

        .. note:: get_molecule returns 'molecules' and likewise
            all methods return a dictionary whose key is the plural
            of the method name. This is quite consistent through the
            API except for that one because it is an acronym

        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("atc_class", query, params=params)

    def get_binding_site(self, limit=20, offset=0, filters=None):
        """Target binding site definition"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("binding_site", query, params=params)

    def get_biotherapeutic(self, limit=20, offset=0, filters=None):
        """Biotherapeutic molecules, which includes HELM notation and sequence data"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("biotherapeutic", query, params=params)

    def get_cell_line(self, limit=20, offset=0, filters=None):
        """Cell line information"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("cell_line", query, params=params)

    def search_chembl_id_lookup(self, query, limit=20, offset=0):
        """Look up ChEMBL Id entity type"""
        params = {"limit": limit, "offset": offset}
        return self._search("chembl_id_lookup", query, params=params)

    def get_chembl_id_lookup(self, query=None, limit=20, offset=0, filters=None):
        """Look up ChEMBL Id entity type"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("chembl_id_lookup", query, params=params)

    def get_compound_record(self, query=None, limit=20, offset=0, filters=None):
        """Occurence of a given compound in a spcecific document"""
        params = {"limit": limit, "offset": offset, "filters": filters}

        return self._get_this_service("compound_record", query, params=params)

    def get_compound_structural_alert(self, query=None, limit=20, offset=0, filters=None):
        """Indicates certain anomaly in compound structure"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("compound_structural_alert", query, params=params)

    def search_document(self, query, limit=20, offset=0):
        """Document/Dataset from which Assays have been extracted"""
        params = {"limit": limit, "offset": offset}
        return self._search("document", query, params=params)

    def get_document(self, query=None, limit=20, offset=0, filters=None):
        """Document/Dataset from which Assays have been extracted"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("document", query, params=params)

    def get_document_similarity(self, query=None, limit=20, offset=0, filters=None):
        """Provides documents similar to a given one"""

        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("document_similarity", query, params=params)

    def get_document_term(self, query=None, limit=20, offset=0, filters=None):
        """Provides keywords extracted from a document using the TextRank algorithm"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("document_term", query, params=params)

    def get_approved_drugs(self, max_phase=4, maxdrugs=1000000):
        """Return all approved drugs

        :param  max_phase: 4 by default for approved drugs.

        """
        filters = "development_phase__exact={}".format(max_phase)
        data = self.get_drug(filters=filters, limit=maxdrugs)
        return data

    def get_drug(self, query=None, limit=20, offset=0, filters=None):
        """Approved drugs information, icluding (but not limited to) applicants, patent numbers and research codes"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("drug", query, params=params)

    def get_drug_indication(self, query=None, limit=20, offset=0, filters=None):
        """Joins drugs with diseases providing references to relevant sources"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("drug_indication", query, params=params)

    def get_go_slim(self, query=None, limit=20, offset=0, filters=None):
        """GO slim ontology"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("go_slim", query, params=params)

    def get_mechanism(self, query=None, limit=20, offset=0, filters=None):
        """Mechanism of action information for FDA-approved drugs"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("mechanism", query, params=params)

    def get_metabolism(self, query=None, limit=20, offset=0, filters=None):
        """Metabolic pathways with references"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("metabolism", query, params=params)

    def search_molecule(self, query, limit=20, offset=0):
        params = {"limit": limit, "offset": offset}
        return self._search("molecule", query, params=params)

    def get_molecule(self, query=None, limit=20, offset=0, filters=None):
        """Returns some molecules

        :param limit: number of molecules to retrieve
        :param offset: molecules to ignore before retrieving molecules.
        :return: a dictionary with keys *page_meta* and *molecules*.

        There are 1,800,000 molecules (Jan 2019). You can only retrieve
        1,000 molecule at most using the *limit* parameter. With a loop
        you can retrieve molecules in some range.

        ::

            c.get_molecule('QFFGVLORLPOAEC-SNVBAGLBSA-N')
            c.get_molecule("CC(=O)Oc1ccccc1C(=O)O")


        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("molecule", query, params=params)

    def get_molecule_form(self, query=None, limit=20, offset=0, filters=None):
        """Relationships between molecule parents and salts


        >>> s.get_molecule_form("CHEMBL2")['molecule_forms']
        [{'is_parent': 'True',
          'molecule_chembl_id': 'CHEMBL2',
          'parent_chembl_id': 'CHEMBL2'},
         {'is_parent': 'False',
          'molecule_chembl_id': 'CHEMBL1558',
          'parent_chembl_id': 'CHEMBL2'},
         {'is_parent': 'False',
          'molecule_chembl_id': 'CHEMBL1347191',
          'parent_chembl_id': 'CHEMBL2'}]
        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("molecule_form", query, params=params)

    def get_organism(self, query=None, limit=20, offset=0, filters=None):
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("organism", query, params=params)

    def search_protein_class(self, query, limit=20, offset=0):
        params = {"limit": limit, "offset": offset}
        return self._search("protein_class", query, params=params)

    def get_protein_class(self, query=None, limit=20, offset=0, filters=None):
        """Protein family classification of TargetComponents"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("protein_class", query, params=params)

    def get_substructure(self, structure, limit=20, offset=0, filters=None):
        """Molecule substructure search


        :param structure: provide a valid / existing substructure in
            SMILE format to look for in all molecules:
        :return: list of molecules corresponding to the search

        ::

            >>> from bioservices import ChEMBL
            >>> c = ChEMBL()
            >>> res = c.get_substructure("CC(=O)Oc1ccccc1C(=O)O")

        Other examples::

            # Substructure search for against ChEMBL using aspirin
            # SMILES string
            c.get_substructure("CC(=O)Oc1ccccc1C(=O)O")

            # Substructure search for against ChEMBL using aspirin
            # CHEMBL_ID
            c.get_substructure("CHEMBL25")

            # Substructure search for against ChEMBL using aspirin
            # InChIKey
            c.get_substructure("BSYNRYMUTXBXSQ-UHFFFAOYSA-N")


        The 'Substructure' and 'Similarity' web service resources allow
        for the chemical content of ChEMBL to be searched. Similar to the
        other resources, these search based resources except filtering, paging
        and ordering arguments. These methods accept SMILES, InChI Key and
        molecule ChEMBL_ID as arguments and in the case of similarity searches
        an additional identity cut-off is needed. Some example molecule searches
        are provided in the table below.

        Searching with InChI key is only possible for InChI keys found in the
        ChEMBL database. The system does not try and convert InChI key to a
        chemical representation.

        """
        # we use quote to formqt the SMILE/InChiKey for a URL parsing
        structure = quote(structure)
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("substructure/{}".format(structure), query, params=params)

    def get_similarity(self, structure, similarity=80, limit=20, offset=0, filters=None):
        """Molecule similarity search

        :param structure: provide a valid / existing substructure in
            SMILE format to look for in all molecules:
        :param similarity: must be an integer greater than 70 and
            less than 100
        :return: list of **molecules** corresponding to the search

        ::

            >>> from bioservices import ChEMBL
            >>> c = ChEMBL()
            >>> res = c.get_similarity("CC(=O)Oc1ccccc1C(=O)O", 80)
            >>> res['molecules']


        Here are more examples::

            # Similarity (80% cut off) search for against ChEMBL using
            # aspirin SMILES string
            c.get_similarity("CC(=O)Oc1ccccc1C(=O)O") # 80 by default

            # Similarity (80% cut off) search for against ChEMBL using
            # aspirin CHEMBL_ID
            c.get_similarity("CHEMBL25")

            # Similarity (80% cut off) search for against ChEMBL
            # using aspirin InChI Key
            c.get_similarity("BSYNRYMUTXBXSQ-UHFFFAOYSA-N")

        The 'Substructure' and 'Similarity' web service resources allow for the
        chemical content of ChEMBL to be searched. Similar to the other resources, these
        search based resources except filtering, paging and ordering arguments. These
        methods accept SMILES, InChI Key and molecule ChEMBL_ID as arguments and in the
        case of similarity searches an additional identity cut-off is needed. Some
        example molecule searches are provided in the table below.

        Searching with InChI key is only possible for InChI keys found in the
        ChEMBL database. The system does not try and convert InChI key to a chemical
        representation.
        """
        # we use quote to formqt the SMILE/InChiKey for a URL parsing
        structure = quote(structure)

        assert isinstance(similarity, int)
        assert similarity >= 70 and similarity <= 100, "similarity must be in the range [70, 100]"
        params = {"limit": limit, "offset": offset, "filters": filters}
        query = None
        return self._get_this_service("similarity/{}/{}".format(structure, similarity), query, params=params)

    def get_source(self, query=None, limit=20, offset=0, filters=None):
        """Document/Dataset source"""
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("source", query, params=params)

    def search_target(self, query, limit=20, offset=0):
        """Targets (protein and non-protein) defined in Assay"""
        params = {"limit": limit, "offset": offset}
        return self._search("target", query, params=params)

    def get_target(self, query=None, limit=20, offset=0, filters=None):
        """Targets (protein and non-protein) defined in Assay


        >>> from bioservices import *
        >>> s = ChEMBL(verbose=False)
        >>> resjson = s.get_targetd('CHEMBL240')

        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("target", query, params=params)

    def get_target_component(self, query=None, limit=20, offset=0, filters=None):
        """Target sequence information (A Target may have 1 or more sequences)

        ::

            res = c.get_target_component(1)
            res['sequence']

        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("target_component", query, params=params)

    def get_target_prediction(self, query=None, limit=20, offset=0, filters=None):
        """Predictied binding of a molecule to a given biological target


        ::

                >>> res = c.get_target_prediction(1)
                >>> res['molecule_chembl_id']
                'CHEMBL2'
        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("target_prediction", query, params=params)

    def get_target_relation(self, query=None, limit=20, offset=0, filters=None):
        """Describes relations between targets


        ::

            >>> c.get_target_relation('CHEMBL261')
            {'related_target_chembl_id': 'CHEMBL2095180',
             'relationship': 'SUBSET OF',
             'target_chembl_id': 'CHEMBL261'}


        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("target_relation", query, params=params)

    def get_tissue(self, query=None, limit=20, offset=0, filters=None):
        """Tissue classification


        c.get_tissue(filters=['pref_name__contains=cervix'])

        """
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("tissue", query, params=params)

    def get_xref_source(self, query=None, limit=20, offset=0, filters=None):
        params = {"limit": limit, "offset": offset, "filters": filters}
        return self._get_this_service("xref_source", query, params=params)

    def get_image(self, query, dimensions=500, format="png", save=True, view=True, engine="indigo"):
        """Get the image of a given compound in PNG png format.

        :param str query: a valid compound ChEMBLId or a list/tuple
            of valid compound ChEMBLIds.
        :param format: png, svg. json not supported
        :param int dimensions: size of image in pixels.
            An integer z (:math:`1 \leq z \leq 500`)
        :param save:
        :param view:
        :param engine: Defaults to rdkit. can be rdkit or indigo
        :param bool view: show the image if set to True.
        :return: the path (list of paths) used to save the figure (figures) (different from Chembl API)

        .. plot::
            :include-source:
            :width: 50%

            >>> from pylab import imread, imshow
            >>> from bioservices import *
            >>> s = ChEMBL(verbose=False)
            >>> res = s.get_image(31863)
            >>> imshow(imread(res['filenames'][0]))

        .. todo:: ignorecoords option
        """
        # NOTE: not async requests here.
        self.devtools.check_range(dimensions, 1, 500)
        self.devtools.check_param_in_list(engine, ["rdkit", "indigo"])
        self.devtools.check_param_in_list(format, ["png", "svg"])
        queries = self.devtools.to_list(query)

        res = {"filenames": [], "images": [], "chemblids": []}
        for query in queries:
            req = "image/{}".format(query)
            params = {"engine": engine, "format": format, "dimensions": dimensions}
            target_data = self.services.http_get(req, frmt=None, params=params)

            file_out = os.getcwd()
            if format == "png":
                file_out += "/%s.png" % query
                with open(file_out, "wb") as thisfile:
                    thisfile.write(bytes(target_data))
            elif format == "svg":
                file_out += "/%s.svg" % query
                with open(file_out, "w") as thisfile:
                    thisfile.write(target_data)
            self.services.logging.info("saved to %s" % file_out)

            fout = file_out
            res["chemblids"].append(query)
            res["filenames"].append(fout)
            res["images"].append(target_data)
        if view:
            webbrowser.open(res["filenames"][0])
        return res

    def get_status(self):
        """Return version of the DB and number of entries

        Returns the number of entries for activities, compound_records,
        distinct_compounds (molecule), publications (document), targets,
        etc...

        .. seealso:: :meth:`get_status_resources`
        """
        return self.services.http_get("status?format=json")

    def get_status_resources(self):
        """Return number of entries for all resources

        .. note:: not in the ChEMBL API.

        .. versionchanged:: 1.7.3 (removed target_prediction and document_term)
        """

        def _local_get(this):
            params = {"limit": 1, "offset": 0}
            return self.services.http_get("{}?format=json".format(this), params=params)["page_meta"]["total_count"]

        data = {}
        for this in [
            "activity",
            "assay",
            "atc_class",
            "cell_line",
            "binding_site",
            "biotherapeutic",
            "chembl_id_lookup",
            "compound_record",
            "compound_structural_alert",
            "document",
            "document_similarity",
            "drug",
            "drug_indication",
            "go_slim",
            "mechanism",
            "metabolism",
            "molecule",
            "molecule_form",
            "protein_class",
            "source",
            "target",
            "target_component",
            "target_relation",
            "tissue",
        ]:
            self.services.logging.info("Looking at {}".format(this))
            try:
                data[this] = _local_get(this)
            except:
                self.services.logging.warning("{} resources seems down".format(this))
        return data

    def order_by(self, data, name, ascending=True):
        """Ordering data

        we use same API as ChEMBL API using the double underscore
        to indicate a hierarchy in the dictionary. So to access to
        d['a']['b'], we use a__b as the input **name** parameter.
        We only allows 3 levels e.g., a__b__c

        ::


            data = c.get_molecules()
            data1 = c.order_by(data, 'molecule_chembl_id')
            data2 = c.order_by(data, 'molecule_properties__alogp')

        .. note:: the ChEMBL API allows for ordering but we do not use
            that API. Instead, we provide this generic function.
        """

        # FIXME sorry no time for a better solution
        # we allow only 3 levels using 3 if
        if name.count("__") == 0:
            data = sorted(data, key=lambda k: k[name], reverse=not ascending)
        elif name.count("__") == 1:
            n1, n2 = name.split("__")
            data = sorted(data, key=lambda k: k[n1][n2], reverse=not ascending)
        elif name.count("__") == 2:
            n1, n2, n3 = name.split("__")
            data = sorted(data, key=lambda k: k[n1][n2][n3], reverse=not ascending)
        else:
            raise NotImplementedError(
                """Please submit a issue on https://github.com/cokelaer/bioservices to allow this level or ordering together will your code example."""
            )
        return data

    def compounds2accession(self, compounds):
        """For each compound, identifies the target and corresponding UniProt
        accession number

        This is not part of ChEMBL API

        ::

            # we recommend to use cache if you use this method regularly
            c = Chembl(cache=True)
            drugs = c.get_approved_drugs()

            # to speed up example
            drugs = drugs[0:20]
            IDs = [x['molecule_chembl_id] for x in drugs]

            c.compounds2accession(IDs)

        """
        # we jump from compounds to targets through activities
        # Here this is a one to many mapping so we initialise a default
        # dictionary.
        from collections import defaultdict

        compound2target = defaultdict(set)

        filter = "molecule_chembl_id__in={}"

        if isinstance(compounds, list):
            pass
        else:
            compounds = list(compounds)

        for compound in tqdm(compounds):
            # FIXME could get activities by bunch using
            # ",".join(compounds[i:i+10) for example
            activities = self.get_activity(filters=filter.format(compound))
            # get target ChEMBL IDs from activities
            for act in activities:
                compound2target[act["molecule_chembl_id"]].add(act["target_chembl_id"])

        # What we need is to get targets for all targets found in the previous
        # step. For each compound/drug there are hundreds of targets though. And
        # we will call the get_target for each list of hundreds targets. This
        # will take forever. Instead, because there are *only* 12,000 targets,
        # let us download all of them ! This took about 4 minutes on this test but
        # if you use the cache, next time it will be much much quicker. This is
        # not down at the activities level because there are too many entries
        targets = self.get_target(limit=-1)

        # identifies all target chembl id to easily retrieve the entry later on
        target_names = [target["target_chembl_id"] for target in targets]

        # retrieve all uniprot accessions for all targets of each compound
        for compound, targs in compound2target.items():
            accessions = set()
            for target in targs:
                index = target_names.index(target)
                accessions = accessions.union([comp["accession"] for comp in targets[index]["target_components"]])
            compound2target[compound] = accessions

        return compound2target



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2020 - EBI-EMBL - Institut Pasteur
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
# $Id$
"""Interface to the PDB web Service (New API Jan 2021).

.. topic:: What is PDB ?

    :URL: http://www.rcsb.org/pdb/
    :REST: http://search.rcsb.org/#search-api

    .. highlights::

        An Information Portal to Biological Macromolecular Structures

        -- PDB home page, Jan 2021


"""
from bioservices.services import REST


__all__ = ["PDB"]


class PDB:
    """Interface to `PDB <http://search.rcsb.org/>`_ service (new API Jan 2021)

    With the new API, one method called :meth:`~bioservices.pdb.PDB.search` is
    provided by PDB. To perform a search you need to define a query. Here is an
    example

    .. doctest::

        >>> from bioservices import PDB
        >>> s = PDB()
        >>> query = {"query":
        ...              {"type": "terminal",
        ...               "service": "text",
        ...               "parameters": {
        ...                 "value": "thymidine kinase"
        ...                 }
        ...             },
        ...          "return_type": "entry"}
        >>> res = s.search(query, return_type=return_type)


    .. note:: as of December 2020, a new API has be set up by PDB.
        some prevous functionalities such as return list of Ligand are not
        supported anymore (Jan 2021). However, many more powerful searches as
        available. I encourage everyone to look at the PDB page for complex
        examples: http://search.rcsb.org/#examples

    As mentionnaed above, the PDB service provide one method called search available in
    :meth:`~bioservices.pdb.PDB.search`. We will not cover all the power and
    capability of this search function. User should refer to the official PDB help
    for that. Yet, given examples from PDB should all work with this method.

    When possible, we will add convenient aliases function in this class. For
    now we have for example the :meth:`~bioservices.pdb.PDB.get_current_ids` and
    :meth:`~bioservices.pdb.PDB.get_similarity_sequence` that users may find useful.

    The main idea behind the PDB API is to create queries that can access to
    different type of services. A query will need to at least two keys:

    - **query**
    - **return_type**

    Consider this basic example that searches for the text *thymidine kinase*::

        {
          "query": {
            "type": "terminal",
            "service": "text",
            "parameters": {
              "value": "thymidine kinase"
            }
          },
          "return_type": "entry"
        }

    Here the query is defined by a **query** and a **return_type** indeed. The
    return type is a simple value such as **entry**. The query itself is
    composed of 3 pairs of key/value. Here we have the type service and
    parameters as defined below.

    The query can have several fields:

    - **type**: the clause type can be either **terminal** or **group**

        - **terminal**: performs an atomic search operation, e.g. searches
          for a particular value in a particular field.
        - **group**: wraps other terminal or group nodes and is
          used to combine multiple queries in a logical fashion.

    - **service**:

        - **text**: linguistic searches against textual annotations.
        - **sequence**: uses MMSeq2 to perform sequence matching searches (blast-like).
          following targets that are available:

          - pdb_protein_sequence,
          - pdb_dna_sequence,
          - pdb_na_sequence
        - **seqmotif**: performs short motif searches against nucleotide or protein
          sequences using 3 different inputs:

          - simple (e.g., CXCXXL)
          - prosite (e.g., C-X-C-X(2)-[LIVMYFWC])
          - regex (e.g., CXCX{2}[LIVMYFWC])
        - **structure**: searches matching a global 3D shape of assemblies
          or chains of a given entry (identified by PDB ID), in either strict
          (strict_shape_match) or relaxed (relaxed_shape_match) modes
        - strucmotif: Performs structural motif searches on all available PDB structures.
        - chemical: queries of small-molecule constituents of PDB structures,
          based on chemical formula and chemical structure. Queries for matching and similar
          chemical structures can be performed using SMILES and InChI descriptors
          as search targets.

          - graph-strict: atom type, formal charge, bond order, atom and bond chirality,
            aromatic assignment are used as matching criteria for this search type.
          - graph-relaxed: atom type, formal charge and bond order are used as
            matching criteria for this search type.
          - graph-relaxed-stereo: atom type, formal charge, bond order, atom
            and bond chirality are used as matching criteria for this search
            type.
          - fingerprint-similarity: Tanimoto similarity is used as the matching criteria

    Concerning the **return_type** key, it can be one of :

    - entry: a list of PDB IDs.
    - assembly: list of PDB IDs appended with assembly IDs in the format of
      a [pdb_id]-[assembly_id], corresponding to biological assemblies.
    - polymer_entity: list of PDB IDs appended with entity IDs in the format
      of a [pdb_id]_[entity_id], corresponding to polymeric molecular entities.
    - non_polymer_entity: list of PDB IDs appended with entity IDs in the
      format of a [pdb_id]_[entity_id], corresponding to non-polymeric entities (or ligands).
    - polymer_instance: list of PDB IDs appended with asym IDs in the format
      of a [pdb_id].[asym_id], corresponding to instances of certain polymeric
      molecular entities, also known as chains.

    **Optional arguments**

    There are many optional arguments. Let us see a couple of them. Pagination can be
    set (default is 10 entries) using the **request_options** (optional) key.
    Consider this query example::

        {
          "query": {
            "type": "terminal",
            "service": "text",
            "parameters": {
                "attribute": "rcsb_polymer_entity.formula_weight",
                "operator": "greater",
                "value": 500
            }
          },
          "request_options": {
            "pager": {
              "start": 0,
              "rows": 100
            }
          },
          "return_type": "polymer_entity"
        }

    Here, the query searches for the polymer_entity that have a formula weight
    above 500. Withe request_options pager set to 100, we will get the first 100
    hits.

    To return all hits, set this field in the request_options::

        "return_all_hits": true

    Coming back at the first basic example, we can reuse it to illustrate how to
    refine the search using attribute and operators::

        {
          "query": {
            "type": "terminal",
            "service": "text",
            "parameters": {
              "value": "thymidine kinase",
              "attribute": "exptl.method",
              "operator": "exact_match",
            }
          },
          "return_type": "entry"
        }

    All valid combo of operators and attributes can be found
    here: http://search.rcsb.org/search-attributes.html

    For instance, in the example above only in, exact_match and exists can be
    used with exptl.method attribute. This is not checked in bioservices.

    Sorting is determined by the sort object in the request_options context.
    It allows you to add one or more sorting conditions to control the order of
    the search result hits. The sort operation is defined on a per field level, with
    special field name for score to sort by score (the default)<

    By default sorting is done in descending order ("desc"). The sort can be
    reversed by setting direction property to "asc". This example demonstrates how
    to sort the search results by release date::

        {
          "query": {
            "type": "terminal",
            "service": "text",
            "parameters": {
              "attribute": "struct.title",
              "operator": "contains_phrase",
              "value": "\"hiv protease\""
            }
          },
          "request_options": {
            "sort": [
              {
                "sort_by": "rcsb_accession_info.initial_release_date",
                "direction": "desc"
              }
            ]
          },
          "return_type": "entry"
        }

    Again, many more complex examples can be found on PDB page.
    """

    _url = "https://search.rcsb.org/rcsbsearch/v2/query"

    def __init__(self, verbose=False, cache=False):
        """.. rubric:: Constructor

        :param bool verbose: prints informative messages (default is off)

        """
        self.services = REST(name="PDB", verbose=verbose, cache=cache, url_defined_later=True)
        self.services.url = PDB._url

    def search(self, query, request_options=None, request_info=None, return_type=None):
        """search request represented as a JSON object.

        This is the only function in PDB API. You should be able
        to perform any valid PDB searches here (see the
        :class:`bioservices.pdb.PDB` documentation for details.
        Note, however, that we have aliases methods in BioServices that will be
        added on demand for common searches.

        :param str query: the search expression. Can be omitted if, instead of IDs retrieval,
            facets or count operation should be performed. In this case the request must be
            configured via the request_options context.
        :param str request_options: (optional) controls various aspects of the search request
            including pagination, sorting, scoring and faceting.
        :param str request_info: additional information about the query, e.g.
            query_id. (optional)
        :param str return_type: type of results to return.
        :return: json results

        You must define a query as defined in the PDB web page. For example the
        following query search for macromolecular PDB entities that share 90% sequence
        identity with GTPase HRas protein from Gallus gallus (Chicken)::

            query = {
              "query": {
                "type": "terminal",
                "service": "sequence",
                "parameters": {
                  "evalue_cutoff": 1,
                  "identity_cutoff": 0.9,
                  "target": "pdb_protein_sequence",
                  "value": "MTEYKLVVVGAGGVGKSALTIQLIQNHFVDEYDPTIEDSYRKQVVIDGETCLLDILDTAGQEEYSAMRDQYMRTGEGFLCVFAINNTKSFEDIHQYREQIKRVKDSDDVPMVLVGNKCDLPARTVETRQAQDLARSYGIPYIETSAKTRQGVEDAFYTLVREIRQHKLRKLNPPDESGPGCMNCKCVIS"
                }
              },
              "request_options": {
                "scoring_strategy": "sequence"
              },
              "return_type": "polymer_entity"
            }

        What is important is that the dictionary called **query** contains 2
        compulsary keys namely **query** and **return_type**. The two other optional
        keys are **request_options** and **return_info**

        You would then call the PDB search as follows::

            from bioservices import PDB
            p = PDB()
            results = p.search(query)

        Now, in BioServices, you can also decompose the query as follows::

            query = {
                "type": "terminal",
                "service": "sequence",
                "parameters": {
                  "evalue_cutoff": 1,
                  "identity_cutoff": 0.9,
                  "target": "pdb_protein_sequence",
                  "value": "MTEYKLVVVGAGGVGKSALTIQLIQNHFVDEYDPTIEDSYRKQVVIDGETCLLDILDTAGQEEYSAMRDQYMRTGEGFLCVFAINNTKSFEDIHQYREQIKRVKDSDDVPMVLVGNKCDLPARTVETRQAQDLARSYGIPYIETSAKTRQGVEDAFYTLVREIRQHKLRKLNPPDESGPGCMNCKCVIS"
                }}
            request_options =  { "scoring_strategy": "sequence"}
            return_type= "polymer_entity"

        and then use PDB search again::

            from bioservices import PDB
            p = PDB()
            results = p.search(query, request_options=request_options, return_type=return_type)

        or even simpler for the Pythonic lovers::

            results = p.search(**query)


        """
        if "query" in query:
            pass
        else:
            query = {"query": query}
            if request_options:
                query["request_options"] = request_options
            if request_info:
                query["request_info"] = request_info
            if return_type:
                query["return_type"] = return_type
        if "return_type" not in query:  # pragma: no cover
            raise ValueError("Your query must have a return_type key")
        print
        res = self.services.http_post("", frmt="json", json=query)
        return res

    def get_current_ids(self):
        """Get a list of all current PDB IDs."""

        # first query returns 10 entries by default

        request_options = {"return_all_hits": True}

        # second requests all entries
        res = self.search(
            query={"type": "terminal", "service": "text"},
            request_options=request_options,
            return_type="entry",
        )

        identifiers = [x["identifier"] for x in res["result_set"]]
        return identifiers

    def get_similarity_sequence(self, seq):
        """Search of seauence similarity search with protein sequence

        seq = "VLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTTKTYFPHFDLSHGSAQVKGHGKKVADALTAVAHVDDMPNAL"
        results = p.get_similarity_sequence(seq)

        """
        res = self.search(
            {
                "query": {
                    "type": "terminal",
                    "service": "sequence",
                    "parameters": {"target": "pdb_protein_sequence", "value": seq},
                },
                "return_type": "polymer_entity",
            }
        )
        return res



from bioservices import UniProt


class Mapping(object):
    """


    Could use unichem, uniprot, kegg, chembldb
    Could be HGNC


    """

    def __init__(self):
        self._uniprot = UniProt()
        self._mapping["uniprot"] = self._uniprot._mapping
        self.databases = {}

    def map(self):
        pass

    def search(self):
        pass
